# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Juniper JCL generator."""

from typing import Optional, Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr, policy, summarizer


# generic error class
class Error(aclgenerator.Error):
    pass


class JuniperTermPortProtocolError(Error):
    pass


class TcpEstablishedWithNonTcpError(Error):
    pass


class JuniperDuplicateTermError(Error):
    pass


class UnsupportedFilterError(Error):
    pass


class PrecedenceError(Error):
    pass


class JuniperIndentationError(Error):
    pass


class JuniperNextIpError(Error):
    pass


class JuniperMultipleTerminatingActionError(Error):
    pass


class JuniperFragmentInV6Error(Error):
    pass


class Config:
    """Config allows a configuration to be assembled easily.

    Configurations are automatically indented following Juniper's style.
    A textual representation of the config can be extracted with str().

    Attributes:
      indent: The number of leading spaces on the current line.
      tabstop: The number of spaces to indent for a new level.
      lines: the text lines of the configuration.
    """

    def __init__(self, indent: int = 0, tabstop: int = 4):
        self.indent = indent
        self._initial_indent = indent
        self.tabstop = tabstop
        self.lines = []

    def __str__(self):
        if self.indent != self._initial_indent:
            raise JuniperIndentationError(
                'Expected indent %d but got %d' % (self._initial_indent, self.indent)
            )
        return '\n'.join(self.lines)

    def Append(self, line: str, verbatim=False):
        """Append one line to the configuration.

        Args:
          line: The string to append to the config.
          verbatim: append line without adjusting indentation. Default False.
        Raises:
          JuniperIndentationError: If the indentation would be further left
            than the initial indent.  e.g. too many close braces.
        """
        if verbatim:
            self.lines.append(line)
            return

        if line.endswith('}'):
            self.indent -= self.tabstop
            if self.indent < self._initial_indent:
                raise JuniperIndentationError('Too many close braces.')
        spaces = ' ' * self.indent
        self.lines.append(spaces + line.strip())
        if not line.find('/*') >= 0 and line.find('*/') >= 0:
            self.indent -= 1
            if self.indent < self._initial_indent:
                raise JuniperIndentationError('Too many close comments.')
        if not line.find('*/') >= 0 and line.find('/*') >= 0:
            self.indent += 1
        if line.endswith(' {'):
            self.indent += self.tabstop


class Term(aclgenerator.Term):
    """Representation of an individual Juniper term.

      This is mostly useful for the __str__() method.
    Attributes:
      term: The term object from policy.
      term_type: String indicating type of term, inet, inet6 icmp etc.
      enable_dsmo: Boolean to enable dsmo.
      noverbose: Boolean to disable verbosity.
      filter_direction: Enum indicating the direction of the filter on an
        interface e.g. INGRESS.
      interface_type: Enum indicating the type of interface filter will be applied
        e.g. LOOPBACK.
      filter_type: String indicating the type of the filter, which may be mixed.
    """

    _PLATFORM = 'juniper'
    _DEFAULT_INDENT = 12
    ACTIONS = {
        'accept': 'accept',
        'deny': 'discard',
        'reject': 'reject',
        'next': 'next term',
        'reject-with-tcp-rst': 'reject tcp-reset',
        'encapsulate': 'encapsulate',
        'port-mirror': 'port-mirror',
    }

    # the following lookup table is used to map between the various types of
    # filters the juniper generator can render.  As new differences are
    # encountered, they should be added to this table.  Accessing members
    # of this table looks like:
    #  self._TERM_TYPE('inet').get('saddr') -> 'source-address'
    #
    # it's critical that the members of each filter type be the same, that is
    # to say that if _TERM_TYPE.get('inet').get('foo') returns something,
    # _TERM_TYPE.get('inet6').get('foo') must return the inet6 equivalent.
    _TERM_TYPE = {
        'inet': {
            'addr': 'address',
            'saddr': 'source-address',
            'daddr': 'destination-address',
            'protocol': 'protocol',
            'protocol-except': 'protocol-except',
            'tcp-est': 'tcp-established',
        },
        'inet6': {
            'addr': 'address',
            'saddr': 'source-address',
            'daddr': 'destination-address',
            'protocol': 'next-header',
            'protocol-except': 'next-header-except',
            'tcp-est': 'tcp-established',
        },
        'bridge': {
            'addr': 'ip-address',
            'saddr': 'ip-source-address',
            'daddr': 'ip-destination-address',
            'protocol': 'ip-protocol',
            'protocol-except': 'ip-protocol-except',
            'tcp-est': 'tcp-flags "(ack|rst)"',
        },
        'ethernet-switching': {
            'addr': 'ip-address',
            'saddr': 'ip-source-address',
            'daddr': 'ip-destination-address',
            'protocol': 'ip-protocol',
            'protocol-except': 'ip-protocol-except',
            'tcp-est': 'tcp-established',
        },
    }

    def __init__(
        self,
        term: policy.Term,
        term_type: str,
        enable_dsmo: bool,
        noverbose: bool,
        filter_direction: Optional[str] = None,
        interface_type: Optional[str] = None,
        filter_type: Optional[str] = None,
    ):
        super().__init__(term)
        self.term = term
        self.term_type = term_type
        self.enable_dsmo = enable_dsmo
        self.noverbose = noverbose
        # Filter direction and interface type are needed in juniperevo sub-class for IPv6 filters.
        self.filter_direction = filter_direction
        self.interface_type = interface_type
        self.filter_type = filter_type

        if self._PLATFORM != 'msmpc':
            if term_type not in self._TERM_TYPE:
                raise ValueError(f'Unknown Filter Type: {term_type}')
            if 'hopopt' in self.term.protocol:
                loc = self.term.protocol.index('hopopt')
                self.term.protocol[loc] = 'hop-by-hop'
            if 'hopopt' in self.term.protocol_except:
                loc = self.term.protocol_except.index('hopopt')
                self.term.protocol_except[loc] = 'hop-by-hop'

            # some options need to modify the actions
            self.extra_actions = []

    def __str__(self):
        config = Config(indent=self._DEFAULT_INDENT)
        from_str = []
        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.term_type == 'inet6' and 'icmp' in self.term.protocol) or (
            self.term_type == 'inet'
            and ('icmpv6' in self.term.protocol or 'icmp6' in self.term.protocol)
        ):
            if self.filter_type != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_PROTO.substitute(
                        term=self.term.name, proto=', '.join(self.term.protocol), af=self.term_type
                    )
                )
            return ''

        # comment
        # this deals just fine with multi line comments, but we could probably
        # output them a little cleaner; do things like make sure the
        # len(output) < 80, etc. Note, if 'noverbose' is set for the filter, skip
        # all comment processing.
        if self.term.owner and not self.noverbose:
            self.term.comment.append(f'Owner: {self.term.owner}')
        if self.term.comment and not self.noverbose:
            config.Append('/*')
            for comment in self.term.comment:
                for line in comment.split('\n'):
                    config.Append(f"** {line}")
            config.Append('*/')

        # Term verbatim output - this will skip over normal term creation
        # code.  Warning generated from policy.py if appropriate.
        if self.term.verbatim:
            for next_term in self.term.verbatim:
                if next_term[0] == self._PLATFORM:
                    config.Append(str(next_term[1]), verbatim=True)
            return str(config)

        # Helper for per-address-family keywords.
        family_keywords = self._TERM_TYPE.get(self.term_type)

        # option
        # this is going to be a little ugly b/c there are a few little messed
        # up options we can deal with.
        if self.term.option:
            for opt in [str(x) for x in self.term.option]:
                # there should be a better way to search the array of protocols
                if opt.startswith('sample'):
                    self.extra_actions.append('sample')

                # only append tcp-established for option established when
                # tcp is the only protocol, otherwise other protos break on juniper
                elif opt.startswith('established'):
                    if self.term.protocol == ['tcp']:
                        if 'tcp-established;' not in from_str:
                            from_str.append(f"{family_keywords['tcp-est']};")

                # if tcp-established specified, but more than just tcp is included
                # in the protocols, raise an error
                elif opt.startswith('tcp-established'):
                    flag = f"{family_keywords['tcp-est']};"
                    if self.term.protocol == ['tcp']:
                        if flag not in from_str:
                            from_str.append(flag)
                    else:
                        raise TcpEstablishedWithNonTcpError(
                            'tcp-established can only be used with tcp protocol in term %s'
                            % self.term.name
                        )
                elif opt.startswith('rst'):
                    from_str.append('tcp-flags "rst";')
                elif opt.startswith('initial') and 'tcp' in self.term.protocol:
                    from_str.append('tcp-initial;')
                elif opt.startswith('first-fragment'):
                    from_str.append('first-fragment;')
                elif opt.startswith('not-syn-ack'):
                    from_str.append('tcp-flags "!(syn&ack)";')

                # we don't have a special way of dealing with this, so we output it and
                # hope the user knows what they're doing.
                else:
                    from_str.append(f'{opt};')

        # if the term is inactive we have to set the prefix
        if self.term.inactive:
            term_prefix = 'inactive:'
        else:
            term_prefix = ''

        # term name
        config.Append(f'{term_prefix} term {self.term.name} {{')

        # The "filter" keyword is not compatible with from or then
        if self.term.filter_term:
            config.Append(f'filter {self.term.filter_term};')
            config.Append('}')  # end term accept-foo-to-bar { ... }
            return str(config)

        # a default action term doesn't have any from { clause
        has_match_criteria = (
            self.term.address
            or self.term.dscp_except
            or self.term.dscp_match
            or self.term.destination_address
            or self.term.destination_port
            or self.term.destination_prefix
            or self.term.destination_prefix_except
            or self.term.encapsulate
            or self.term.ether_type
            or self.term.flexible_match_range
            or self.term.forwarding_class
            or self.term.forwarding_class_except
            or self.term.fragment_offset
            or self.term.hop_limit
            or self.term.next_ip
            or self.term.port
            or self.term.precedence
            or self.term.protocol
            or self.term.protocol_except
            or self.term.source_address
            or self.term.source_port
            or self.term.source_prefix
            or self.term.source_prefix_except
            or self.term.traffic_type
            or self.term.ttl
        )

        if has_match_criteria:
            config.Append('from {')

            term_af = self.AF_MAP.get(self.term_type)

            # address
            address = self.term.GetAddressOfVersion('address', term_af)
            if self.enable_dsmo:
                address = summarizer.Summarize(address)

            if address:
                config.Append('%s {' % family_keywords['addr'])
                for addr in address:
                    for comment in self._Comment(addr):
                        config.Append(f'{comment}')
                    if self.enable_dsmo:
                        config.Append('%s/%s;' % summarizer.ToDottedQuad(addr, nondsm=True))
                    else:
                        config.Append(f'{addr};')
                config.Append('}')
            elif self.term.address:
                if self.filter_type != 'mixed':
                    logging.warning(
                        self.NO_AF_LOG_ADDR.substitute(term=self.term.name, af=self.term_type)
                    )
                return ''

            # source address

            src_addr = self.term.GetAddressOfVersion('source_address', term_af)
            src_addr_ex = self.term.GetAddressOfVersion('source_address_exclude', term_af)
            if self.enable_dsmo:
                src_addr = summarizer.Summarize(src_addr)
                src_addr_ex = summarizer.Summarize(src_addr_ex)
            else:
                src_addr, src_addr_ex = self._MinimizePrefixes(src_addr, src_addr_ex)

            if src_addr:
                config.Append('%s {' % family_keywords['saddr'])
                for addr in src_addr:
                    for comment in self._Comment(addr):
                        config.Append(f'{comment}')
                    if self.enable_dsmo:
                        config.Append('%s/%s;' % summarizer.ToDottedQuad(addr, nondsm=True))
                    else:
                        config.Append(f'{addr};')
                for addr in src_addr_ex:
                    for comment in self._Comment(addr, exclude=True):
                        config.Append(f'{comment}')
                    if self.enable_dsmo:
                        config.Append('%s/%s except;' % summarizer.ToDottedQuad(addr, nondsm=True))
                    else:
                        config.Append(f'{addr} except;')
                config.Append('}')
            elif self.term.source_address:
                if self.filter_type != 'mixed':
                    logging.warning(
                        self.NO_AF_LOG_ADDR.substitute(
                            term=self.term.name, direction='source', af=self.term_type
                        )
                    )
                return ''

            # destination address
            dst_addr = self.term.GetAddressOfVersion('destination_address', term_af)
            dst_addr_ex = self.term.GetAddressOfVersion('destination_address_exclude', term_af)
            if self.enable_dsmo:
                dst_addr = summarizer.Summarize(dst_addr)
                dst_addr_ex = summarizer.Summarize(dst_addr_ex)
            else:
                dst_addr, dst_addr_ex = self._MinimizePrefixes(dst_addr, dst_addr_ex)

            if dst_addr:
                config.Append('%s {' % family_keywords['daddr'])
                for addr in dst_addr:
                    for comment in self._Comment(addr):
                        config.Append(f'{comment}')
                    if self.enable_dsmo:
                        config.Append('%s/%s;' % summarizer.ToDottedQuad(addr, nondsm=True))
                    else:
                        config.Append(f'{addr};')
                for addr in dst_addr_ex:
                    for comment in self._Comment(addr, exclude=True):
                        config.Append(f'{comment}')
                    if self.enable_dsmo:
                        config.Append('%s/%s except;' % summarizer.ToDottedQuad(addr, nondsm=True))
                    else:
                        config.Append(f'{addr} except;')
                config.Append('}')
            elif self.term.destination_address:
                if self.filter_type != 'mixed':
                    logging.warning(
                        self.NO_AF_LOG_ADDR.substitute(
                            term=self.term.name, direction='destination', af=self.term_type
                        )
                    )
                return ''

            # forwarding-class
            if self.term.forwarding_class:
                config.Append(
                    f'forwarding-class {self._Group(self.term.forwarding_class, lc=False)}'
                )

            # forwarding-class-except
            if self.term.forwarding_class_except:
                config.Append(
                    'forwarding-class-except %s'
                    % self._Group(self.term.forwarding_class_except, lc=False)
                )

            # source prefix <except> list
            if self.term.source_prefix or self.term.source_prefix_except:
                config.Append('source-prefix-list {')
                for pfx in self.term.source_prefix:
                    config.Append(f"{pfx};")
                for epfx in self.term.source_prefix_except:
                    config.Append(f"{epfx} except;")
                config.Append('}')

            # destination prefix <except> list
            if self.term.destination_prefix or self.term.destination_prefix_except:
                config.Append('destination-prefix-list {')
                for pfx in self.term.destination_prefix:
                    config.Append(f"{pfx};")
                for epfx in self.term.destination_prefix_except:
                    config.Append(f"{epfx} except;")
                config.Append('}')

            # Only generate ttl if inet, inet6 uses hop-limit instead.
            if self.term.ttl and self.term_type == 'inet':
                config.Append(f'ttl {self.term.ttl};')

            # protocol
            if self.term.protocol:
                # Juniper has deprecated the use of icmpv6 in favor of icmp6.
                # https://github.com/aerleon/aerleon/issues/323
                protocol = self.term.protocol[:]
                if 'icmpv6' in self.term.protocol:
                    loc = protocol.index('icmpv6')
                    protocol[loc] = 'icmp6'
                # both are supported on JunOS, but only icmp6 is supported
                # on SRX loopback stateless filter
                config.Append(f"{family_keywords['protocol']} {self._Group(protocol)}")

            # protocol
            if self.term.protocol_except:
                # same as above
                protocol_except = self.term.protocol_except[:]
                if 'icmpv6' in self.term.protocol_except:
                    loc = protocol_except.index('icmpv6')
                    protocol_except[loc] = 'icmp6'
                config.Append(
                    f"{family_keywords['protocol-except']} {self._Group(protocol_except)}"
                )

            # port
            if self.term.port:
                config.Append(f'port {self._Group(self.term.port)}')

            # source port
            if self.term.source_port:
                config.Append(f'source-port {self._Group(self.term.source_port)}')

            # destination port
            if self.term.destination_port:
                config.Append(f'destination-port {self._Group(self.term.destination_port)}')

            # append any options beloging in the from {} section
            for next_str in from_str:
                config.Append(next_str)

            # packet length
            if self.term.packet_length:
                config.Append(f'packet-length {self.term.packet_length};')

            # fragment offset
            if self.term.fragment_offset:
                config.Append(f'fragment-offset {self.term.fragment_offset};')

            # icmp-types
            icmp_types = ['']
            if self.term.icmp_type:
                icmp_types = self.NormalizeIcmpTypes(
                    self.term.icmp_type, self.term.protocol, self.term_type
                )
            if icmp_types != ['']:
                config.Append(f'icmp-type {self._Group(icmp_types)}')
            if self.term.icmp_code:
                config.Append(f'icmp-code {self._Group(self.term.icmp_code)}')
            if self.term.ether_type:
                config.Append(f'ether-type {self._Group(self.term.ether_type)}')

            if self.term.traffic_type:
                config.Append(f'traffic-type {self._Group(self.term.traffic_type)}')

            if self.term.precedence:
                # precedence may be a single integer, or a space separated list
                policy_precedences = set()
                # precedence values may only be 0 through 7
                for precedence in self.term.precedence:
                    if int(precedence) in range(0, 8):
                        policy_precedences.add(precedence)
                    else:
                        raise PrecedenceError(
                            'Precedence value %s is out of bounds in %s'
                            % (precedence, self.term.name)
                        )
                config.Append(f'precedence {self._Group(sorted(policy_precedences))}')

            # DSCP Match
            if self.term.dscp_match:
                if self.term_type == 'inet6':
                    config.Append(f"traffic-class [ {' '.join(self.term.dscp_match)} ];")
                else:
                    config.Append(f"dscp [ {' '.join(self.term.dscp_match)} ];")

            # DSCP Except
            if self.term.dscp_except:
                if self.term_type == 'inet6':
                    config.Append(f"traffic-class-except [ {' '.join(self.term.dscp_except)} ];")
                else:
                    config.Append(f"dscp-except [ {' '.join(self.term.dscp_except)} ];")

            if self.term.hop_limit:
                # Only generate a hop-limit if inet6, inet4 has not hop-limit.
                if self.term_type == 'inet6':
                    config.Append(f'hop-limit {self.term.hop_limit};')

            # flexible-match
            if self.term.flexible_match_range:
                config.Append('flexible-match-range {')
                for fm_opt in self.term.flexible_match_range:
                    config.Append(f'{fm_opt[0]} {fm_opt[1]};')
                config.Append('}')

            config.Append('}')  # end from { ... }

        ####
        # ACTIONS go below here
        ####

        # If the action is only one line, include it in the same line as "then "
        # statement.
        # For example, if the action is only accept, it should be:
        # "then accept;" rather than:
        # "then {
        #     accept;
        # }"
        #
        self.CheckTerminatingAction()
        unique_actions = set(self.extra_actions)
        if not self.term.routing_instance:
            unique_actions.update(self.term.action)
        if self.term.encapsulate:
            unique_actions.add('encapsulate')
        if len(unique_actions) <= 1:
            for action in [
                self.term.logging,
                self.term.routing_instance,
                self.term.counter,
                self.term.policer,
                self.term.qos,
                self.term.loss_priority,
                self.term.dscp_set,
                self.term.next_ip,
                self.term.traffic_class_count,
                self.term.port_mirror,
            ]:
                if action:
                    try:
                        unique_actions.update(action)
                    except TypeError:
                        unique_actions.add(action)
                    if len(unique_actions) > 1:
                        break

        if len(unique_actions) == 1:
            # b/21795531: Juniper device treats a set of IPv4 actions differently
            # than any other actions.
            # For example, if the term is in IPv4 and the action is only discard,
            # it should be:
            # "then {
            #     discard;
            # }" rather than:
            # "then discard;"
            current_action = self.ACTIONS.get(unique_actions.pop(), 'next_ip')
            if (
                self.term_type == 'inet'
                and current_action in ['discard', 'reject', 'reject tcp-reset']
            ) or (self.term_type == 'inet6' and current_action in ['reject', 'reject tcp-reset']):
                config.Append('then {')
                config.Append(f'{current_action};')
                config.Append('}')
            elif current_action == 'next_ip':
                self.NextIpCheck(self.term.next_ip, self.term.name)
                config.Append('then {')
                if self.term.next_ip[0].version == 4:
                    config.Append(f'next-ip {self.term.next_ip[0]!s};')
                else:
                    config.Append(f'next-ip6 {self.term.next_ip[0]!s};')
                config.Append('}')
            elif current_action == 'encapsulate':
                config.Append('then {')
                config.Append(f'encapsulate {self.term.encapsulate!s};')
                config.Append('}')
            else:
                config.Append(f'then {current_action};')
        elif len(unique_actions) > 1:
            config.Append('then {')
            # logging
            if self.term.logging:
                for log_target in self.term.logging:
                    if str(log_target) == 'local':
                        config.Append('log;')
                    else:
                        config.Append('syslog;')

            if self.term.routing_instance:
                config.Append(f'routing-instance {self.term.routing_instance};')

            if self.term.counter:
                config.Append(f'count {self.term.counter};')

            if self.term.traffic_class_count:
                config.Append(f'traffic-class-count {self.term.traffic_class_count};')

            oid_length = 128
            if self.term.policer:
                config.Append(f'policer {self.term.policer};')
                if len(self.term.policer) > oid_length:
                    logging.warning(
                        'WARNING: %s is longer than %d bytes. Due to '
                        'limitation in JUNOS, OIDs longer than %dB can '
                        'cause SNMP timeout issues.',
                        self.term.policer,
                        oid_length,
                        oid_length,
                    )

            if self.term.qos:
                config.Append(f'forwarding-class {self.term.qos};')

            if self.term.port_mirror:
                config.Append('port-mirror;')
            if self.term.loss_priority:
                config.Append(f'loss-priority {self.term.loss_priority};')
            if self.term.next_ip:
                self.NextIpCheck(self.term.next_ip, self.term.name)
                if self.term.next_ip[0].version == 4:
                    config.Append(f'next-ip {self.term.next_ip[0]!s};')
                else:
                    config.Append(f'next-ip6 {self.term.next_ip[0]!s};')
            if self.term.encapsulate:
                config.Append(f'encapsulate {self.term.encapsulate!s};')
            for action in self.extra_actions:
                config.Append(f"{action};")

            # If there is a routing-instance defined, skip reject/accept/etc actions.
            if not self.term.routing_instance:
                for action in self.term.action:
                    config.Append(f"{self.ACTIONS.get(action)};")

            # DSCP SET
            if self.term.dscp_set:
                if self.term_type == 'inet6':
                    config.Append(f'traffic-class {self.term.dscp_set};')
                else:
                    config.Append(f'dscp {self.term.dscp_set};')

            config.Append('}')  # end then{...}

        config.Append('}')  # end term accept-foo-to-bar { ... }

        return str(config)

    @staticmethod
    def NextIpCheck(next_ip: list[Union[nacaddr.IPv4, nacaddr.IPv6]], term_name: str):
        if len(next_ip) > 1:
            raise JuniperNextIpError(
                f'The following term has more than one next IP value: {term_name}'
            )
        if next_ip[0].num_addresses > 1:
            raise JuniperNextIpError(
                f'The following term has a subnet instead of a host: {term_name}'
            )

    def CheckTerminatingAction(self):
        action = set(self.term.action)
        if self.term.encapsulate:
            action.add(self.term.encapsulate)
        if self.term.routing_instance:
            action.add(self.term.routing_instance)
        if len(action) > 1:
            raise JuniperMultipleTerminatingActionError(
                f'The following term has multiple terminating actions: {self.term.name}'
            )

    def _MinimizePrefixes(
        self,
        include: list[Union[nacaddr.IPv4, nacaddr.IPv6]],
        exclude: list[Union[nacaddr.IPv4, nacaddr.IPv6]],
    ):
        """Calculate a minimal set of prefixes for Juniper match conditions.

        Args:
          include: Iterable of nacaddr objects, prefixes to match.
          exclude: Iterable of nacaddr objects, prefixes to exclude.
        Returns:
          A tuple (I,E) where I and E are lists containing the minimized
          versions of include and exclude, respectively.  The order
          of each input list is preserved.
        """
        # Remove any included prefixes that have EXACT matches in the
        # excluded list.  Excluded prefixes take precedence on the router
        # regardless of the order in which the include/exclude are applied.
        exclude_set = set(exclude)
        include_result = [ip for ip in include if ip not in exclude_set]

        # Every address match condition on a Juniper firewall filter
        # contains an implicit "0/0 except" or "0::0/0 except".  If an
        # excluded prefix is not contained within any less-specific prefix
        # in the included set, we can elide it.  In other words, if the
        # next-less-specific prefix is the implicit "default except",
        # there is no need to configure the more specific "except".
        #
        # TODO(kbrint): this could be made more efficient with a Patricia trie.
        exclude_result = []
        for exclude_prefix in exclude:
            for include_prefix in include_result:
                if exclude_prefix.subnet_of(include_prefix):
                    exclude_result.append(exclude_prefix)
                    break

        return include_result, exclude_result

    def _Comment(
        self,
        addr: Union[nacaddr.IPv4, nacaddr.IPv6, summarizer.DSMNet],
        exclude: bool = False,
        line_length: int = 132,
    ):
        """Returns address comment field if it exists.

        Args:
          addr: nacaddr.IPv4 object (?)
          exclude: bool - address excludes have different indentations
          line_length: integer - this is the length to which a comment will be
            truncated, no matter what.  ie, a 1000 character comment will be
            truncated to line_length, and then split.  if 0, the whole comment
            is kept. the current default of 132 is somewhat arbitrary.

        Returns:
          List of strings.

        Notes:
          This method tries to intelligently split long comments up.  if we've
          managed to summarize 4 /32's into a /30, each with a nacaddr text field
          of something like 'foobar N', normal concatination would make the
          resulting rendered comment look in mondrian like

                             source-address {
                                 ...
                                 1.1.1.0/30; /* foobar1, foobar2, foobar3, foo
          bar4 */

          b/c of the line splitting at 80 chars.  this method will split the
          comments at word breaks and make the previous example look like

                             source-address {
                                  ....
                                  1.1.1.0/30; /* foobar1, foobar2, foobar3,
                                              ** foobar4 */
          much cleaner.
        """
        rval = []
        if self.noverbose:
            return rval
        # indentation, for multi-line comments, ensures that subsquent lines
        # are correctly alligned with the first line of the comment.
        indentation = 0
        if exclude:
            # len('1.1.1.1/32 except;') == 21
            indentation = 21 + self._DEFAULT_INDENT
        else:
            # len('1.1.1.1/32;') == 14
            indentation = 14 + self._DEFAULT_INDENT

        # length_eol is the width of the line; b/c of the addition of the space
        # and the /* characters, it needs to be a little less than the actual width
        # to keep from wrapping
        length_eol = 77 - indentation

        if isinstance(addr, (nacaddr.IPv4, nacaddr.IPv6, summarizer.DSMNet)):
            if addr.text:
                if line_length == 0:
                    # line_length of 0 means that we don't want to truncate the comment.
                    line_length = len(addr.text)

                # There should never be a /* or */, but be safe and ignore those
                # comments
                if addr.text.find('/*') >= 0 or addr.text.find('*/') >= 0:
                    logging.warning('Malformed comment [%s] ignoring', addr.text)
                else:
                    text = addr.text[:line_length]

                    comment = ' /*'
                    while text:
                        # split the line
                        if len(text) > length_eol:
                            new_length_eol = text[:length_eol].rfind(' ')
                            if new_length_eol <= 0:
                                new_length_eol = length_eol
                        else:
                            new_length_eol = length_eol

                        # what line am I gunna output?
                        line = f"{comment} {text[:new_length_eol].strip()}"
                        # truncate what's left
                        text = text[new_length_eol:]
                        # setup the comment and indentation for the next go-round
                        comment = f"{' ' * indentation}**"

                        rval.append(line)

                    rval[-1] += ' */'
        else:
            # should we be paying attention to any other addr type?
            logging.warning('Ignoring non IPv4 or IPv6 address: %s', addr)
        return rval

    def _Group(self, group: list[str], lc: bool = True):
        """If 1 item return it, else return [ item1 item2 ].

        Args:
          group: a list.  could be a list of strings (protocols) or a list of
                 tuples (ports)
          lc: return a lower cased result for text.  Default is True.

        Returns:
          rval: a string surrounded by '[' and '];' if len(group) > 1
                or with just ';' appended if len(group) == 1
        """

        def _FormattedGroup(el: list[str], lc: bool = True):
            """Return the actual formatting of an individual element.

            Args:
              el: either a string (protocol) or a tuple (ports)
              lc: return lower cased result for text.  Default is True.

            Returns:
              string: either the lower()'ed string or the ports, hyphenated
                      if they're a range, or by itself if it's not.
            """
            if isinstance(el, str):
                if lc:
                    return el
                else:
                    return el.lower()
            elif isinstance(el, int):
                return str(el)
            # type is a tuple below here
            elif el[0] == el[1]:
                return '%d' % el[0]
            else:
                return '%d-%d' % (el[0], el[1])

        if len(group) > 1:
            rval = f"[ {' '.join([_FormattedGroup(x) for x in group])} ];"
        else:
            rval = f"{_FormattedGroup(group[0])};"
        return rval


class Juniper(aclgenerator.ACLGenerator):
    """JCL rendering class.

      This class takes a policy object and renders the output into a syntax
      which is understood by juniper routers.

    Attributes:
      pol: policy.Policy object
    """

    _PLATFORM = 'juniper'
    _DEFAULT_PROTOCOL = 'ip'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'bridge', 'ethernet-switching', 'mixed'))
    _TERM = Term
    SUFFIX = '.jcl'

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'address',
            'restrict_address_family',
            'counter',
            'destination_prefix',
            'destination_prefix_except',
            'dscp_except',
            'dscp_match',
            'dscp_set',
            'encapsulate',
            'ether_type',
            'filter_term',
            'flexible_match_range',
            'forwarding_class',
            'forwarding_class_except',
            'fragment_offset',
            'hop_limit',
            'icmp_code',
            'logging',
            'loss_priority',
            'next_ip',
            'owner',
            'packet_length',
            'policer',
            'port',
            'port_mirror',
            'precedence',
            'protocol_except',
            'qos',
            'routing_instance',
            'source_prefix',
            'source_prefix_except',
            'traffic_type',
            'traffic_class_count',
            'ttl',
        }
        supported_sub_tokens.update(
            {
                'option': {
                    'established',
                    'first-fragment',
                    'is-fragment',
                    # TODO(sneakywombat): add all options to lex.
                    '.*',  # make ArbitraryOptions work, yolo.
                    'sample',
                    'tcp-established',
                    'tcp-initial',
                    'inactive',
                    'not-syn-ack',
                }
            }
        )
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int):
        self.juniper_policies = []
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            # Check for the position independent options and remove them from
            # the list.
            interface_specific = 'not-interface-specific' not in filter_options[1:]
            enable_dsmo = 'enable_dsmo' in filter_options[1:]
            noverbose = 'noverbose' in filter_options[1:]
            filter_enhanced_mode = 'filter_enhanced_mode' in filter_options[1:]

            filter_direction = None
            if 'ingress' in filter_options[1:]:
                filter_direction = 'ingress'
            elif 'egress' in filter_options[1:]:
                filter_direction = 'egress'
            interface_type = None
            if 'physical' in filter_options[1:]:
                interface_type = 'physical'
            elif 'loopback' in filter_options[1:]:
                interface_type = 'loopback'

            if not interface_specific:
                filter_options.remove('not-interface-specific')
            if enable_dsmo:
                filter_options.remove('enable_dsmo')
            if filter_enhanced_mode:
                filter_options.remove('filter_enhanced_mode')

            # default to inet4 filters
            filter_type = 'inet'
            if len(filter_options) > 1:
                filter_type = filter_options[1]

            if filter_type == 'mixed':
                filter_types_to_process = ['inet', 'inet6']
            else:
                filter_types_to_process = [filter_type]

            for term_filter_type in filter_types_to_process:
                filter_name_suffix = ''
                # If mixed filter_type, will append 4 or 6 to the filter name
                if len(filter_types_to_process) > 1:
                    if term_filter_type == 'inet':
                        filter_name_suffix = '4'
                    if term_filter_type == 'inet6':
                        filter_name_suffix = '6'

                term_names = set()
                new_terms = []
                for term in terms:
                    # Ignore if the term is for a different AF
                    if (
                        term.restrict_address_family
                        and term.restrict_address_family != term_filter_type
                    ):
                        continue

                    # if inactive is set, deactivate the term and remove the option.
                    if 'inactive' in term.option:
                        term.inactive = True
                        term.option.remove('inactive')

                    term.name = self.FixTermLength(term.name)

                    if term.name in term_names:
                        raise JuniperDuplicateTermError(
                            f'You have multiple terms named: {term.name}'
                        )
                    term_names.add(term.name)

                    term = self.FixHighPorts(term, af=term_filter_type)
                    if not term:
                        continue

                    if 'is-fragment' in term.option and term_filter_type == 'inet6':
                        raise JuniperFragmentInV6Error(
                            f'The term {term.name} uses "is-fragment" but is a v6 policy.'
                        )

                    new_terms.append(
                        self._TERM(
                            term,
                            term_filter_type,
                            enable_dsmo,
                            noverbose,
                            filter_direction,
                            interface_type,
                            filter_type=filter_type,
                        )
                    )

                self.juniper_policies.append(
                    (
                        header,
                        filter_name + filter_name_suffix,
                        term_filter_type,
                        interface_specific,
                        filter_enhanced_mode,
                        new_terms,
                    )
                )

    def __str__(self):
        config = Config()

        for (
            header,
            filter_name,
            filter_type,
            interface_specific,
            filter_enhanced_mode,
            terms,
        ) in self.juniper_policies:
            # add the header information
            config.Append('firewall {')
            config.Append('family %s {' % filter_type)
            config.Append('/*')

            # we want the acl to contain id and date tags, but p4 will expand
            # the tags here when we submit the generator, so we have to trick
            # p4 into not knowing these words.  like taking c-a-n-d-y from a
            # baby.
            for line in aclgenerator.AddRepositoryTags('** '):
                config.Append(line)
            config.Append('**')

            for comment in header.comment:
                for line in comment.split('\n'):
                    config.Append(f"** {line}")
            config.Append('*/')

            config.Append('replace: filter %s {' % filter_name)
            if interface_specific:
                config.Append('interface-specific;')
            if filter_enhanced_mode:
                config.Append('enhanced-mode;')

            for term in terms:
                term_str = str(term)
                if term_str:
                    config.Append(term_str, verbatim=True)

            config.Append('}')  # filter { ... }
            config.Append('}')  # family inet { ... }
            config.Append('}')  # firewall { ... }

        return f"{config!s}\n"



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pcap filter generator.

This generate a pcap packet filter expression that either:
  1) Matches (i.e., captures), the packets that match the ACCEPT clauses
     specified in a given policy, or
  2) Matches the packets that match opposite of that, i.e., the DENY or REJECT
     clauses.
Support tcp flags matching and icmptypes, including ipv6/icmpv6, but not much
else past the standard addres, port, and protocol conditions.

Note that this is still alpha and will likely require more testing prior to
having more confidence in it.

Stolen liberally from packetfilter.py.
"""


from typing import Union

from absl import logging

from aerleon.lib import aclgenerator
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term


class Error(aclgenerator.Error):
    """Base error class."""


class UnsupportedActionError(Error):
    """Raised when we see an unsupported action."""


class UnsupportedTargetOptionError(Error):
    """Raised when we see an unsupported option."""


class Term(aclgenerator.Term):
    """Generate pcap filter to match a policy term."""

    _PLATFORM = 'pcap'
    _ACTION_TABLE = {
        'accept': '',
        'deny': '',
        'reject': '',
        'next': '',
    }

    _TCP_FLAGS_TABLE = {
        'syn': 'tcp-syn',
        'ack': 'tcp-ack',
        'fin': 'tcp-fin',
        'rst': 'tcp-rst',
        'urg': 'tcp-urg',
        'psh': 'tcp-push',
        'all': '(tcp-syn|tcp-ack|tcp-fin|tcp-rst|tcp-urg|tcp-push)',
        'none': '(tcp-syn&tcp-ack&tcp-fin&tcp-rst&tcp-urg&tcp-push)',
    }

    _PROTO_TABLE = {
        'ah': 'proto \\ah',
        'esp': 'proto \\esp',
        'icmp': 'proto \\icmp',
        'icmpv6': 'icmp6',
        'ip': 'proto \\ip',
        'ip6': 'ip6',
        'igmp': 'proto \\igmp',
        'igrp': 'igrp',
        'pim': 'proto \\pim',
        'tcp': 'proto \\tcp',
        'udp': 'proto \\udp',
        # bpf supports "\vrrp", but some winpcap version dont' recognize it,
        # so use the IANA protocol number for it:
        'vrrp': 'proto 112',
        'hopopt': 'ip6 protochain 0',
    }

    def __init__(
        self, term: Term, filter_name: str, af: str = 'inet', direction: str = ''
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in packetfilter.
          filter_name: The name of the filter chan to attach the term to.
          af: Which address family ('inet' or 'inet6') to apply the term to.
          direction: Direction of the flow.

        Raises:
          aclgenerator.UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.options = []
        self.default_action = 'deny'
        self.af = af
        self.direction = direction

    def __str__(self) -> str:
        """Render config output from this term object."""
        conditions = []

        # if terms does not specify action, use filter default action
        if not self.term.action:
            self.term.action[0].value = self.default_action
        if str(self.term.action[0]) not in self._ACTION_TABLE:
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s %s'
                % ('\n', self.term.name, self.term.action[0], 'action not currently supported.')
            )

        # source address
        term_saddrs = self._CheckAddressAf(self.term.source_address)
        if not term_saddrs:
            if self.af != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_ADDR.substitute(
                        term=self.term.name, direction='source', af=self.af
                    )
                )
            return ''

        conditions.append(
            self._GenerateAddrStatement(term_saddrs, self.term.source_address_exclude)
        )

        # destination address
        term_daddrs = self._CheckAddressAf(self.term.destination_address)
        if not term_daddrs:
            if self.af != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_ADDR.substitute(
                        term=self.term.name, direction='destination', af=self.af
                    )
                )
            return ''

        conditions.append(
            self._GenerateAddrStatement(term_daddrs, self.term.destination_address_exclude)
        )

        # protocol
        if self.term.protocol_except:
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s'
                % ('\n', self.term.name, 'protocol_except logic not currently supported.')
            )
        conditions.append(self._GenerateProtoStatement(self.term.protocol))

        conditions.append(self._GeneratePortStatement(self.term.source_port, 'src'))
        conditions.append(self._GeneratePortStatement(self.term.destination_port, 'dst'))

        # icmp-type
        icmp_types = ['']
        if self.term.icmp_type:
            if self.term.protocol == ['icmp']:
                af = 'inet'
            elif self.term.protocol == ['icmpv6']:
                af = 'inet6'
            else:
                raise aclgenerator.UnsupportedFilterError(
                    '%s %s %s'
                    % ('\n', self.term.name, 'icmp protocol is not defined or not supported.')
                )
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, self.term.protocol, af)

            if 'icmp' in self.term.protocol:
                conditions.append(self._GenerateIcmpType(icmp_types, self.term.icmp_code))

        # tcp options
        if 'tcp' in self.term.protocol:
            conditions.append(self._GenerateTcpOptions(self.term.option))

        cond = Term.JoinConditionals(conditions, 'and')

        # Note that directionally-based pcap filter requires post-processing to
        # replace 'localhost' with whatever the IP(s) of the local machine happen
        # to be.  This bit of logic ensure there's a placeholder with the
        # appropriate booleans around it.  We also have to check that there exists
        # some form of condition already, else we'll end up with something overly
        # broad like 'dst net localhost' (e.g., 'default-deny').
        if cond and self.direction == 'in':
            cond = Term.JoinConditionals(['dst net localhost', cond], 'and')
        elif cond and self.direction == 'out':
            cond = Term.JoinConditionals(['src net localhost', cond], 'and')

        return cond + '\n'

    def _CheckAddressAf(self, addrs: list[Union[IPv4, IPv6]]) -> list[Union[str, IPv4, IPv6]]:
        """Verify that the requested address-family matches the address's family."""
        if not addrs:
            return ['any']
        if self.af == 'mixed':
            return addrs
        af_addrs = []
        af = self.NormalizeAddressFamily(self.af)
        for addr in addrs:
            if addr.version == af:
                af_addrs.append(addr)
        return af_addrs

    @staticmethod
    def JoinConditionals(condition_list: list[str], operator: str) -> str:
        """Join conditionals using the specified operator.

        Filters out empty elements and blank strings.

        Args:
          condition_list:  a list of str()-able items to join.
          operator:  the join string.

        Returns:
          A string consisting of the joined elements.  If all elements are False
          or whitespace-only, the empty string.
        """
        condition_list = [_f for _f in condition_list if _f]
        condition_list = [str(x).strip(' ') for x in condition_list if str(x).strip()]
        if not condition_list:
            return ''

        op = f' {operator} '
        res = f'({op.join(condition_list)})'
        return res

    def _GenerateAddrStatement(
        self, addrs: list[Union[str, IPv6, IPv4]], exclude_addrs: list[Union[str, IPv6, IPv4]]
    ) -> str:
        addrlist = []
        for d in addrs:
            if d != 'any' and str(d) != '::/0':
                addrlist.append(f'dst net {d}')

        excludes = []
        if exclude_addrs:
            for d in exclude_addrs:
                if d != 'any' and str(d) != '::/0':
                    excludes.append(f'not dst net {d}')
                else:
                    # excluding 'any' doesn't really make sense ...
                    return ''

        if excludes:
            return Term.JoinConditionals(
                [Term.JoinConditionals(addrlist, 'or'), Term.JoinConditionals(excludes, 'or')],
                'and not',
            )
        else:
            return Term.JoinConditionals(addrlist, 'or')

    def _GenerateProtoStatement(self, protocols: list[str]) -> str:
        return Term.JoinConditionals([self._PROTO_TABLE[p] for p in protocols], 'or')

    def _GeneratePortStatement(self, ports: list[tuple[int, int]], direction: str) -> str:
        conditions = []
        # term.destination_port is a list of tuples containing the start and end
        # ports of the port range.  In the event it is a single port, the start
        # and end ports are the same.
        for port_tuple in ports:
            if port_tuple[0] == port_tuple[1]:
                conditions.append(f'{direction} port {port_tuple[0]}')
            else:
                conditions.append(f'{direction} portrange {port_tuple[0]}-{port_tuple[1]}')
        return Term.JoinConditionals(conditions, 'or')

    def _GenerateTcpOptions(self, options: list[str]) -> str:
        opts = [str(x) for x in options]
        tcp_flags_set = []
        tcp_flags_check = []
        for next_opt in opts:
            if next_opt == 'tcp-established':
                tcp_flags_set.append(self._TCP_FLAGS_TABLE['ack'])
                tcp_flags_check.extend([self._TCP_FLAGS_TABLE['ack']])

            else:
                # Iterate through flags table, and create list of tcp-flags to append
                for next_flag in self._TCP_FLAGS_TABLE:
                    if next_opt.find(next_flag) == 0:
                        tcp_flags_check.append(self._TCP_FLAGS_TABLE.get(next_flag))
                        tcp_flags_set.append(self._TCP_FLAGS_TABLE.get(next_flag))

        if tcp_flags_check:
            return '(tcp[tcpflags] & ({}) == ({}))'.format(
                '|'.join(tcp_flags_check),
                '|'.join(tcp_flags_set),
            )
        return ''

    def _GenerateIcmpType(self, icmp_types: list[int], icmp_code: list[int]) -> str:
        rtr_str = ''
        if icmp_types:
            code_strings = ['']
            if icmp_code:
                code_strings = [' and icmp[icmpcode] == %d' % code for code in icmp_code]
            rtr_str = Term.JoinConditionals(
                ['icmp[icmptype] == %d%s' % (x, y) for y in code_strings for x in icmp_types], 'or'
            )
        return rtr_str


class PcapFilter(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object.

    Note that since pcap isn't actually a firewall grammar, this generates a
    filter that only matches matches that which would be accepted by the
    specified policy.
    """

    _PLATFORM = 'pcap'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = '.pcap'
    _TERM = Term

    def __init__(self, *args, **kwargs) -> None:
        """Initialize a PcapFilter generator.

        Takes standard ACLGenerator arguments, as well as an 'invert' kwarg.  If
        this argument is true, the pcap filter will be reversed, such that it
        matches all those packets that would be denied by the specified policy.

        Args:
          *args: Arguments.
          **kwargs: Keyword arguments.

        """
        self._invert = False
        if 'invert' in kwargs:
            self._invert = kwargs['invert']
            del kwargs['invert']
        super().__init__(*args, **kwargs)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'logging', 'icmp_code'}
        supported_tokens -= {'verbatim'}

        supported_sub_tokens.update(
            {
                'action': {'accept', 'deny', 'reject', 'next'},
                'option': {
                    'tcp-established',
                    'established',
                    'syn',
                    'ack',
                    'fin',
                    'rst',
                    'urg',
                    'psh',
                    'all',
                    'none',
                },
            }
        )

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        self.pcap_policies = []
        good_afs = ['inet', 'inet6', 'mixed']
        good_options = ['in', 'out']
        direction = ''

        for header, terms in pol.filters:
            filter_type = None
            filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)

            # ensure all options after the filter name are expected
            for opt in filter_options:
                if opt not in good_afs + good_options:
                    raise UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            if 'in' in filter_options:
                direction = 'in'
            elif 'out' in filter_options:
                direction = 'out'

            # Check for matching af
            for address_family in good_afs:
                if address_family in filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise aclgenerator.UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                good_afs,
                                'in filter options:',
                                filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'mixed'

            # add the terms
            accept_terms = []
            deny_terms = []
            term_names = set()
            for term in terms:
                if term.name in term_names:
                    raise aclgenerator.DuplicateTermError(
                        f'You have a duplicate term: {term.name}'
                    )

                if not term:
                    continue

                if term.action[0] == 'accept':
                    accept_terms.append(self._TERM(term, filter_name, filter_type, direction))
                elif term.action[0] == 'deny' or term.action[0] == 'reject':
                    deny_terms.append(self._TERM(term, filter_name, filter_type, direction))

            self.pcap_policies.append((header, filter_name, filter_type, accept_terms, deny_terms))

    def __str__(self) -> str:
        """Render the output of the PF policy into config."""
        target = []

        for (
            unused_header,
            unused_filter_name,
            unused_filter_type,
            accept_terms,
            deny_terms,
        ) in self.pcap_policies:
            accept = []
            for term in accept_terms:
                term_str = str(term)
                if term_str:
                    accept.append(str(term))
            accept_clause = Term.JoinConditionals(accept, 'and')

            deny = []
            for term in deny_terms:
                term_str = str(term)
                if term_str:
                    deny.append(str(term))
            deny_clause = Term.JoinConditionals(deny, 'and')

            if self._invert:
                target.append(Term.JoinConditionals([deny_clause, accept_clause], 'and not'))
            else:
                target.append(Term.JoinConditionals([accept_clause, deny_clause], 'and not'))

        return '\nor\n'.join(target) + '\n'



# Copyright 2020-2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
"""A module of utilities to work with IP addresses in a faster way."""

import ipaddress
from collections.abc import Iterator
from typing import Union


def exclude_address(
    base_net: ipaddress._BaseNetwork,  # pylint disable=protected-access
    exclude_net: ipaddress._BaseNetwork,  # pylint disable=protected-access
) -> Iterator[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]:
    """
    Function to exclude a subnetwork from another, returning a generator that
    yields all values that correspond to the base network without the exclude
    network.

    This is functionally equivalent to the _BaseNetwork "address_exclude" from the
    `ipaddress` standard library, but is a faster implementation since
    the standard library function is a O(n) operation on the length of the
    netmask of the excluding network, whereas this function is O(1) for all cases.

    Args:
      base_net: an object of type _BaseNetwork, the network that
                contains the exclude network
      exclude_net: an object of type _BaseNetwork, the network
                   that is being removed from the base_net
    Raises:
      ValueError if exclude_net is not completely contained in base_net

    Yields:
      A sequence of IP networks that do not encompass the exclude_net
    """

    if not isinstance(base_net, ipaddress._BaseNetwork):  # pylint disable=protected-access
        raise TypeError(f'{base_net} is not a network object')

    if not isinstance(exclude_net, ipaddress._BaseNetwork):  # pylint disable=protected-access
        raise TypeError(f'{exclude_net} is not a network object')

    if base_net.version != exclude_net.version:
        raise TypeError(f'{base_net} and {exclude_net} are not of the same version')

    if not exclude_net.subnet_of(base_net):  # pytype: disable=attribute-error
        raise ValueError()
    if exclude_net == base_net:
        return

    include_range = (
        base_net.network_address._ip,
        base_net.broadcast_address._ip,
    )  # pylint disable=protected-access # pytype: disable=attribute-error
    exclude_range = (
        exclude_net.network_address._ip,
        exclude_net.broadcast_address._ip,
    )  # pylint disable=protected-access # pytype: disable=attribute-error
    address_class = base_net.network_address.__class__  # pylint disable=protected-access
    if include_range[0] == exclude_range[0]:
        result_start = address_class(exclude_range[1] + 1)
        result_end = address_class(include_range[1])
        yield from ipaddress.summarize_address_range(result_start, result_end)
    elif include_range[1] == exclude_range[1]:
        result_start = address_class(include_range[0])
        result_end = address_class(exclude_range[0] - 1)
        yield from ipaddress.summarize_address_range(result_start, result_end)
    else:
        first_section_start = address_class(include_range[0])
        first_section_end = address_class(exclude_range[0] - 1)
        second_section_start = address_class(exclude_range[1] + 1)
        second_section_end = address_class(include_range[1])
        yield from ipaddress.summarize_address_range(first_section_start, first_section_end)
        yield from ipaddress.summarize_address_range(second_section_start, second_section_end)



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ipset iptables generator.  This is a subclass of Iptables generator.

ipset is a system inside the Linux kernel, which can very efficiently store
and match IPv4 and IPv6 addresses. This can be used to dramatically increase
performace of iptables firewall.

"""

import string
from typing import Any, Union

from aerleon.lib import iptables, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6


class Error(iptables.Error):
    """Base error class."""


class Term(iptables.Term):
    """Single Ipset term representation."""

    _PLATFORM = 'ipset'
    _SET_MAX_LENGTH = 31
    _POSTJUMP_FORMAT = None
    _PREJUMP_FORMAT = None
    _TERM_FORMAT = None
    _COMMENT_FORMAT = string.Template('-A $filter -m comment --comment "$comment"')
    _FILTER_TOP_FORMAT = string.Template('-A $filter')

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # This stores tuples of set name and set contents, keyed by direction.
        # For example:
        # { 'src': ('set_name', [ipaddr object, ipaddr object]),
        #   'dst': ('set_name', [ipaddr object, ipaddr object]) }
        self.addr_sets = {}

    def _CalculateAddresses(
        self,
        src_addr_list: list[Union[IPv4, IPv6]],
        src_addr_exclude_list: list[Union[IPv4, IPv6]],
        dst_addr_list: list[Union[IPv4, IPv6]],
        dst_addr_exclude_list: list[Union[IPv4, IPv6]],
    ) -> tuple[
        list[Union[IPv4, IPv6]],
        list[Union[IPv4, IPv6]],
        list[Union[IPv4, IPv6]],
        list[Union[IPv4, IPv6]],
    ]:
        """Calculates source and destination address list for a term.

        Since ipset is very efficient at matching large number of
        addresses, we never return any exclude addresses. Instead
        least positive match is calculated for both source and destination
        addresses.

        For source and destination address list, three cases are possible.
        First case is when there are no addresses. In that case we return
        _all_ips.
        Second case is when there is strictly one address. In that case,
        we optimize by not generating a set, and it's then the only
        element of returned set.
        Third case is when there are more than one address in a set.
        In that case we generate a set and also return _all_ips. Note the
        difference to the first case where no set is actually generated.

        Args:
          src_addr_list: source address list of the term.
          src_addr_exclude_list: source address exclude list of the term.
          dst_addr_list: destination address list of the term.
          dst_addr_exclude_list: destination address exclude list of the term.

        Returns:
          tuple containing source address list, source address exclude list,
          destination address list, destination address exclude list in
          that order.

        """
        target_af = self.AF_MAP[self.af]
        src_addr_list = self._CalculateAddrList(
            src_addr_list, src_addr_exclude_list, target_af, 'src'
        )
        dst_addr_list = self._CalculateAddrList(
            dst_addr_list, dst_addr_exclude_list, target_af, 'dst'
        )
        return (src_addr_list, [], dst_addr_list, [])

    def _CalculateAddrList(
        self,
        addr_list: list[Union[IPv4, IPv6]],
        addr_exclude_list: list[Any],
        target_af: int,
        direction: str,
    ) -> list[Union[IPv4, IPv6]]:
        """Calculates and stores address list for target AF and direction.

        Args:
          addr_list: address list.
          addr_exclude_list: address exclude list of the term.
          target_af: target address family.
          direction: direction in which address list will be used.

        Returns:
          calculated address list.

        """
        if not addr_list:
            addr_list = [self._all_ips]
        addr_list = [addr for addr in addr_list if addr.version == target_af]
        if addr_exclude_list:
            addr_exclude_list = [
                addr_exclude
                for addr_exclude in addr_exclude_list
                if addr_exclude.version == target_af
            ]
            addr_list = nacaddr.ExcludeAddrs(addr_list, addr_exclude_list)
        if len(addr_list) > 1:
            set_name = self._GenerateSetName(self.term.name, direction)
            self.addr_sets[direction] = (set_name, addr_list)
            addr_list = [self._all_ips]
        return addr_list

    def _GenerateAddressStatement(self, src_addr: IPv4, dst_addr: IPv4) -> tuple[str, str]:
        """Returns the address section of an individual iptables rule.

        See _CalculateAddresses documentation. Three cases are possible here,
        and they map directly to cases in _CalculateAddresses.
        First, there can be no addresses for a direction (value is _all_ips then)
        In that case we return empty string.
        Second there can be stricly one address. In that case we return single
        address match (-s or -d).
        Third case, is when the value is _all_ips but also the set for particular
        direction is present. That's when we return a set match.

        Args:
          src_addr: ipaddr address or network object with source
            address of the rule.
          dst_addr: ipaddr address or network object with destination
            address of the rule.

        Returns:
          tuple containing source and destination address statement, in
          that order.

        """
        src_addr_stmt = ''
        dst_addr_stmt = ''
        if src_addr and dst_addr:
            if src_addr == self._all_ips:
                if 'src' in self.addr_sets:
                    src_addr_stmt = f"-m set --match-set {self.addr_sets['src'][0]} src"
            else:
                src_addr_stmt = '-s %s/%d' % (src_addr.network_address, src_addr.prefixlen)
            if dst_addr == self._all_ips:
                if 'dst' in self.addr_sets:
                    dst_addr_stmt = f"-m set --match-set {self.addr_sets['dst'][0]} dst"
            else:
                dst_addr_stmt = '-d %s/%d' % (dst_addr.network_address, dst_addr.prefixlen)
        return (src_addr_stmt, dst_addr_stmt)

    def _GenerateSetName(self, term_name: str, suffix: str) -> str:
        if self.af == 'inet6':
            suffix += '-v6'
        if len(term_name) + len(suffix) + 1 > self._SET_MAX_LENGTH:
            set_name_max_lenth = self._SET_MAX_LENGTH - len(suffix) - 1
            term_name = term_name[:set_name_max_lenth]
        return f'{term_name}-{suffix}'


class Ipset(iptables.Iptables):
    """Ipset generator."""

    _PLATFORM = 'ipset'
    _SET_TYPE = 'hash:net'
    SUFFIX = '.ips'
    _TERM = Term
    _MARKER_BEGIN = '# begin:ipset-rules'
    _MARKER_END = '# end:ipset-rules'
    _GOOD_OPTIONS = ['nostate', 'abbreviateterms', 'truncateterms', 'noverbose', 'exists']

    def __str__(self) -> str:
        # Actual rendering happens in __str__, so it has to be called
        # before we do set specific part.
        iptables_output = super().__str__()
        output = []
        output.append(self._MARKER_BEGIN)
        for _, _, _, _, terms in self.iptables_policies:
            for term in terms:
                output.extend(self._GenerateSetConfig(term))
        output.append(self._MARKER_END)
        output.append(iptables_output)
        return '\n'.join(output)

    def _GenerateSetConfig(self, term: Term) -> list[str]:
        """Generates set configuration for supplied term.

        Args:
          term: input term.

        Returns:
          string that is configuration of supplied term.

        """
        output = []
        c_str = 'create'
        a_str = 'add'
        if 'exists' in self.filter_options:
            c_str = f"{c_str} -exist"
            a_str = f"{a_str} -exist"
        for direction in sorted(term.addr_sets, reverse=True):
            set_name, addr_list = term.addr_sets[direction]
            set_hashsize = 1 << len(addr_list).bit_length()
            set_maxelem = set_hashsize
            output.append(
                '%s %s %s family %s hashsize %i maxelem %i'
                % (c_str, set_name, self._SET_TYPE, term.af, set_hashsize, set_maxelem)
            )
            for address in addr_list:
                output.append(f'{a_str} {set_name} {address}')
        return output



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""SRX generator."""
# pylint: disable=super-init-not-called


import collections
import copy
import itertools
from typing import Union

from absl import logging

from aerleon.lib import aclgenerator, addressbook, nacaddr, policy

ICMP_TERM_LIMIT = 8
FQDNSUFFIX = '_FQDN'


def JunipersrxList(name, data):
    return f"{name} [ {' '.join(data)} ];"


class Error(aclgenerator.Error):
    """generic error class."""


class UnsupportedFilterError(Error):
    pass


class UnsupportedHeaderError(Error):
    pass


class SRXDuplicateTermError(Error):
    pass


class SRXVerbatimError(Error):
    pass


class SRXOptionError(Error):
    pass


class MixedAddrBookTypesError(Error):
    pass


class ConflictingTargetOptionsError(Error):
    pass


class ConflictingApplicationSetsError(Error):
    pass


class IndentList(list):
    def __init__(self, indent: str, *args, **kwargs):
        self._indent = indent
        super().__init__(*args, **kwargs)

    def IndentAppend(self, size, data):
        self.append(f'{self._indent * size}{data}')


class Term(aclgenerator.Term):
    """Representation of an individual SRX term.

    This is mostly useful for the __str__() method.

    Args:
      obj: a policy.Term object
      filter_options: list of remaining target options (zones)
    """

    ACTIONS = {
        'accept': 'permit',
        'deny': 'deny',
        'reject': 'reject',
        'count': 'count',
        'log': 'log',
        'expresspath': 'services-offload',
        'dscp': 'dscp',
    }

    def __init__(
        self,
        term: policy.Term,
        from_zone: str,
        to_zone: str,
        expresspath: bool = False,
        verbose: bool = True,
    ):
        super().__init__(term)
        self.term = term
        self.from_zone = from_zone
        self.to_zone = to_zone
        self.verbose = verbose
        if expresspath:
            self.term.action = [a.replace('accept', 'expresspath') for a in self.term.action]

    def __str__(self) -> str:
        """Render config output from this term object."""
        ret_str = IndentList(JuniperSRX.INDENT)

        # COMMENTS
        comment_max_width = 68
        if self.term.owner and self.verbose:
            self.term.comment.append(f'Owner: {self.term.owner}')
        comments = aclgenerator.WrapWords(self.term.comment, comment_max_width)
        if comments and comments[0] and self.verbose:
            ret_str.IndentAppend(3, '/*')
            for line in comments:
                ret_str.IndentAppend(3, line)
            ret_str.IndentAppend(3, '*/')

        ret_str.IndentAppend(3, f"policy {self.term.name} {{")
        ret_str.IndentAppend(4, 'match {')
        # SOURCE-ADDRESS
        saddrs = {i.parent_token for i in self.term.source_address}
        saddrs.update({i.parent_token + FQDNSUFFIX for i in self.term.source_fqdn})
        if saddrs:
            ret_str.IndentAppend(5, JunipersrxList('source-address', sorted(saddrs)))
        else:
            ret_str.IndentAppend(5, 'source-address any;')

        # DESTINATION-ADDRESS
        daddrs = {i.parent_token for i in self.term.destination_address}
        daddrs.update({i.parent_token + FQDNSUFFIX for i in self.term.destination_fqdn})
        if daddrs:
            ret_str.IndentAppend(5, JunipersrxList('destination-address', sorted(daddrs)))
        else:
            ret_str.IndentAppend(5, 'destination-address any;')

        # APPLICATION
        if (
            not self.term.source_port
            and not self.term.destination_port
            and not self.term.icmp_type
            and not self.term.protocol
        ):
            ret_str.IndentAppend(5, 'application any;')
        else:
            if hasattr(self.term, 'replacement_application_name'):
                ret_str.IndentAppend(
                    5, f"application {self.term.replacement_application_name}-app;"
                )
            else:
                ret_str.IndentAppend(5, f"application {self.term.name}-app;")

        # DSCP MATCH
        if self.term.dscp_match:
            ret_str.IndentAppend(5, JunipersrxList('dscp', self.term.dscp_match))

        # DSCP EXCEPT
        if self.term.dscp_except:
            ret_str.IndentAppend(5, JunipersrxList('dscp-except', self.term.dscp_except))

        # SOURCE-ZONE
        if self.term.source_zone:
            szone_check = set()
            for szone in self.term.source_zone:
                szone_check.add(szone)
            szone_check = sorted(szone_check)
            ret_str.IndentAppend(5, JunipersrxList('from-zone', szone_check))

        # DESTINATION-ZONE
        if self.term.destination_zone:
            dzone_check = set()
            for dzone in self.term.destination_zone:
                dzone_check.add(dzone)
            dzone_check = sorted(dzone_check)
            ret_str.IndentAppend(5, JunipersrxList('to-zone', dzone_check))

        ret_str.IndentAppend(4, '}')

        # ACTIONS
        for action in self.term.action:
            ret_str.IndentAppend(4, 'then {')

            # VPN target can be only specified when ACTION is accept
            if str(action) == 'accept' and self.term.vpn:
                ret_str.IndentAppend(5, f"{self.ACTIONS.get(str(action))} {{")
                ret_str.IndentAppend(6, 'tunnel {')
                ret_str.IndentAppend(7, f'ipsec-vpn {self.term.vpn[0]};')
                if self.term.vpn[1]:
                    ret_str.IndentAppend(7, f'pair-policy {self.term.vpn[1]};')

                ret_str.IndentAppend(6, '}')
                ret_str.IndentAppend(5, '}')
            else:
                ret_str.IndentAppend(5, f"{self.ACTIONS.get(str(action))};")

            # DSCP SET
            if self.term.dscp_set:
                ret_str.IndentAppend(5, f"dscp {self.term.dscp_set};")

            # LOGGING
            if self.term.logging:
                ret_str.IndentAppend(5, 'log {')
                for log_target in self.term.logging:
                    if str(log_target) == 'log-both':
                        ret_str.IndentAppend(6, 'session-init;')
                        ret_str.IndentAppend(6, 'session-close;')
                    else:
                        if str(action) == 'accept':
                            ret_str.IndentAppend(6, 'session-close;')
                        else:
                            ret_str.IndentAppend(6, 'session-init;')
                ret_str.IndentAppend(5, '}')

            ret_str.IndentAppend(4, '}')

            ret_str.IndentAppend(3, '}')

        return '\n'.join(ret_str)

    def _Group(self, group: list[str]):
        """If 1 item return it, else return [ item1 item2 ].

        Args:
          group: a list.  could be a list of strings (protocols) or a list of
                 tuples (ports)

        Returns:
          rval: a string surrounded by '[' and '];' if len(group) > 1
                or with just ';' appended if len(group) == 1
        """

        def _FormattedGroup(el: list[str]):
            """Return the actual formatting of an individual element.

            Args:
              el: either a string (protocol) or a tuple (ports)

            Returns:
              string: either the lower()'ed string or the ports, hyphenated
                      if they're a range, or by itself if it's not.
            """
            if isinstance(el, str):
                return el.lower()
            elif isinstance(el, int):
                return str(el)
            # type is a tuple below here
            elif el[0] == el[1]:
                return '%d' % el[0]
            else:
                return '%d-%d' % (el[0], el[1])

        if len(group) > 1:
            rval = f"[ {' '.join([_FormattedGroup(x) for x in group])} ];"
        else:
            rval = f"{_FormattedGroup(group[0])};"
        return rval


class JuniperSRX(aclgenerator.ACLGenerator):
    """SRX rendering class.

    This class takes a policy object and renders the output into a syntax
    which is understood by SRX firewalls.

    Args:
      pol: policy.Policy object
    """

    _PLATFORM = 'srx'
    SUFFIX = '.srx'
    _SUPPORTED_AF = {'inet', 'inet6', 'mixed'}
    _ZONE_ADDR_BOOK = 'address-book-zone'
    _GLOBAL_ADDR_BOOK = 'address-book-global'
    _ADDRESSBOOK_TYPES = {_ZONE_ADDR_BOOK, _GLOBAL_ADDR_BOOK}
    _EXPRESSPATH = 'expresspath'
    _NOVERBOSE = 'noverbose'
    _SUPPORTED_TARGET_OPTIONS = {_ZONE_ADDR_BOOK, _GLOBAL_ADDR_BOOK, _EXPRESSPATH, _NOVERBOSE}

    _AF_MAP = {'inet': (4,), 'inet6': (6,), 'mixed': (4, 6)}
    _AF_ICMP_MAP = {'icmp': 'inet', 'icmpv6': 'inet6'}
    INDENT = '    '
    _MAX_HEADER_COMMENT_LENGTH = 71
    # The SRX platform is limited in how many IP addresses can be used in
    # a single policy.
    _ADDRESS_LENGTH_LIMIT = 1023
    # IPv6 are 32 bytes compared to IPv4, this is used as a multiplier.
    _IPV6_SIZE = 4

    def __init__(self, pol: policy.Policy, exp_info: int):
        self.srx_policies = []
        self.addressbook = addressbook.Addressbook()
        self.applications = []
        self.ports = []
        self.from_zone = ''
        self.to_zone = ''
        self.addr_book_type = set()
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'dscp_except',
            'dscp_match',
            'dscp_set',
            'destination_fqdn',
            'destination_zone',
            'logging',
            'option',
            'owner',
            'source_fqdn',
            'source_zone',
            'timeout',
            'verbatim',
            'vpn',
        }

        supported_sub_tokens.update(
            {
                'action': {'accept', 'deny', 'reject', 'count', 'log', 'dscp'},
            }
        )
        del supported_sub_tokens['option']
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int):
        # pylint: disable=attribute-defined-outside-init
        """Transform a policy object into a JuniperSRX object.

        Args:
          pol: policy.Policy object
          exp_info: print a info message when a term is set to expire
                    in that many weeks

        Raises:
          UnsupportedFilterError: An unsupported filter was specified
          UnsupportedHeaderError: A header option exists that is not
                                  understood/usable
          SRXDuplicateTermError: Two terms were found with same name in same filter
          ConflictingTargetOptionsError: Two target options are conflicting
                                         in the header
          MixedAddrBookTypesError: Global and Zone address books in the same policy
          ConflictingApplicationSetsError: When two duplicate named terms have
                                           conflicting application entries
        """
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)

            verbose = True
            if self._NOVERBOSE in filter_options[4:]:
                verbose = False

            # TODO(robankeny): Clean up option section.
            if (
                len(filter_options) < 4
                or filter_options[0] != 'from-zone'
                or filter_options[2] != 'to-zone'
            ):
                raise UnsupportedFilterError(
                    'SRX filter arguments must specify ' 'from-zone and to-zone.'
                )

            # check if to-zone is not a supported target option
            if filter_options[1] in self._SUPPORTED_TARGET_OPTIONS:
                raise UnsupportedFilterError(
                    'to-zone %s cannot be the same as any '
                    'valid SRX target-options' % (filter_options[1])
                )
            else:
                self.from_zone = filter_options[1]

            # check if from-zone is not a supported target option
            if filter_options[3] in self._SUPPORTED_TARGET_OPTIONS:
                raise UnsupportedFilterError(
                    'from-zone %s cannot be the same as any '
                    'valid SRX target-options' % (filter_options[3])
                )
            else:
                self.to_zone = filter_options[3]

            # check if source-zone & destination-zone are only used with global policy
            if filter_options[1] != 'all' or filter_options[3] != 'all':
                for term in terms:
                    if term.source_zone or term.destination_zone:
                        raise UnsupportedFilterError(
                            'Term %s has either source-zone or '
                            'destination-zone which can only be '
                            'used with global policy' % term.name
                        )

            # variables used to collect target-options and set defaults
            filter_type = ''

            # parse srx target options
            extra_options = filter_options[4:]
            if self._ADDRESSBOOK_TYPES.issubset(extra_options):
                raise ConflictingTargetOptionsError(
                    'only one address-book-type can '
                    'be specified per header "%s"' % ' '.join(filter_options)
                )
            else:
                address_book_type = {self._ZONE_ADDR_BOOK, self._GLOBAL_ADDR_BOOK}.intersection(
                    extra_options
                )
                if not address_book_type:
                    address_book_type = {self._GLOBAL_ADDR_BOOK}
                self.addr_book_type.update(address_book_type)
                if len(self.addr_book_type) > 1:
                    raise MixedAddrBookTypesError(
                        'Global and Zone address-book-types cannot ' 'be used in the same policy'
                    )
                if self.from_zone == 'all' and self.to_zone == 'all':
                    if self._ZONE_ADDR_BOOK in self.addr_book_type:
                        raise UnsupportedFilterError(
                            'Zone address books cannot be used ' 'with a global policy.'
                        )
                elif self.from_zone == 'all' or self.to_zone == 'all':
                    raise UnsupportedFilterError(
                        'The zone name all is reserved for ' 'global policies.'
                    )

            if self._EXPRESSPATH in filter_options[4:]:
                self.expresspath = True
            else:
                self.expresspath = False

            for filter_opt in filter_options[4:]:
                # validate address families
                if filter_opt in self._SUPPORTED_AF:
                    if not filter_type:
                        filter_type = filter_opt
                    else:
                        raise ConflictingTargetOptionsError(
                            'only one address family can be '
                            'specified per header "%s"' % ' '.join(filter_options)
                        )
                elif filter_opt in self._SUPPORTED_TARGET_OPTIONS:
                    continue
                else:
                    raise UnsupportedHeaderError(
                        'SRX Generator currently does not support '
                        '%s as a header option "%s"' % (filter_opt, ' '.join(filter_options))
                    )

            # if address-family and address-book-type have not been set then default
            if not filter_type:
                filter_type = 'mixed'

            term_dup_check = set()

            new_terms = []
            self._FixLargePolices(terms, filter_type)
            for term in terms:
                if term.stateless_reply:
                    logging.warning(
                        "WARNING: Term %s in policy %s>%s is a stateless reply "
                        "term and will not be rendered.",
                        term.name,
                        self.from_zone,
                        self.to_zone,
                    )
                    continue
                if {'established', 'tcp-established'}.intersection(term.option):
                    logging.warning(
                        'Skipping established term %s because SRX is stateful.', term.name
                    )
                    continue
                term.name = self.FixTermLength(term.name)
                if term.name in term_dup_check:
                    raise SRXDuplicateTermError(f'You have a duplicate term: {term.name}')
                term_dup_check.add(term.name)

                # SRX address books leverage network token names for IPs.
                # When excluding addresses, we lose those distinct names so we need
                # to create a new unique name based off the term name before excluding.
                if term.source_address_exclude:
                    # If we have a naked source_exclude, we need something to exclude from
                    if not term.source_address:
                        term.source_address = [
                            nacaddr.IP('0.0.0.0/0', term.name.upper(), term.name.upper())
                        ]
                    # Use the term name as the token & parent_token
                    new_src_parent_token = f"{term.name.upper()}_SRC_EXCLUDE"
                    new_src_token = new_src_parent_token
                    for i in term.source_address_exclude:
                        term.source_address = nacaddr.RemoveAddressFromList(term.source_address, i)
                        for j in term.source_address:
                            j.token = new_src_token
                            j.parent_token = new_src_parent_token

                if term.destination_address_exclude:
                    if not term.destination_address:
                        term.destination_address = [
                            nacaddr.IP('0.0.0.0/0', term.name.upper(), term.name.upper())
                        ]
                    new_dst_parent_token = f"{term.name.upper()}_DST_EXCLUDE"
                    new_dst_token = new_dst_parent_token
                    for i in term.destination_address_exclude:
                        term.destination_address = nacaddr.RemoveAddressFromList(
                            term.destination_address, i
                        )
                        for j in term.destination_address:
                            j.token = new_dst_token
                            j.parent_token = new_dst_parent_token

                # SRX policies are controlled by addresses that are used within, so
                # policy can be at the same time inet and inet6.
                if self._GLOBAL_ADDR_BOOK in self.addr_book_type:
                    for zone in self.addressbook._ip_addressbook:
                        for _, ips in sorted(self.addressbook._ip_addressbook[zone].items()):
                            ips = [i for i in ips]
                            if term.source_address == ips:
                                term.source_address = ips
                            if term.destination_address == ips:
                                term.destination_address = ips

                # Filter source_address based on filter_type & add to address book
                if term.source_address:
                    valid_addrs = []
                    for addr in term.source_address:
                        if addr.version in self._AF_MAP[filter_type]:
                            valid_addrs.append(addr)
                    if not valid_addrs:
                        logging.warning(
                            'WARNING: Term %s has 0 valid source IPs, skipping.', term.name
                        )
                        continue
                    term.source_address = valid_addrs
                    if term.source_address:
                        self.addressbook.AddAddresses(self.from_zone, term.source_address)
                if term.source_fqdn:
                    self.addressbook.AddFQDNs(self.from_zone, term.source_fqdn, suffix=FQDNSUFFIX)

                # Filter destination_address based on filter_type & add to address book
                if term.destination_address:
                    valid_addrs = []
                    for addr in term.destination_address:
                        if addr.version in self._AF_MAP[filter_type]:
                            valid_addrs.append(addr)
                    if not valid_addrs:
                        logging.warning(
                            'WARNING: Term %s has 0 valid destination IPs, skipping.', term.name
                        )
                        continue
                    term.destination_address = valid_addrs
                    if term.destination_address:
                        self.addressbook.AddAddresses(self.to_zone, term.destination_address)
                if term.destination_fqdn:
                    self.addressbook.AddFQDNs(
                        self.to_zone, term.destination_fqdn, suffix=FQDNSUFFIX
                    )
                new_term = Term(term, self.from_zone, self.to_zone, self.expresspath, verbose)
                new_terms.append(new_term)

                # Because SRX terms can contain inet and inet6 addresses. We have to
                # have ability to recover proper AF for ICMP type we need.
                # If protocol is empty or we cannot map to inet or inet6 we insert bogus
                # af_type name which will cause new_term.NormalizeIcmpTypes to fail.
                if not term.protocol:
                    icmp_af_type = 'unknown_af_icmp'
                else:
                    icmp_af_type = self._AF_ICMP_MAP.get(term.protocol[0], 'unknown_af_icmp')
                tmp_icmptype = new_term.NormalizeIcmpTypes(
                    term.icmp_type, term.protocol, icmp_af_type
                )
                # NormalizeIcmpTypes returns [''] for empty, convert to [] for eval
                normalized_icmptype = tmp_icmptype if tmp_icmptype != [''] else []
                # rewrites the protocol icmpv6 to icmp6
                if 'icmpv6' in term.protocol:
                    protocol = list(term.protocol)
                    protocol[protocol.index('icmpv6')] = 'icmp6'
                else:
                    protocol = term.protocol
                new_application_set = {
                    'sport': self._BuildPort(term.source_port),
                    'dport': self._BuildPort(term.destination_port),
                    'protocol': protocol,
                    'icmp-type': normalized_icmptype,
                    'timeout': term.timeout,
                }

                for application_set in self.applications:
                    if all(
                        item in list(application_set.items())
                        for item in new_application_set.items()
                    ):
                        new_application_set = ''
                        term.replacement_application_name = application_set['name']
                        break
                    if (
                        term.name == application_set['name']
                        and new_application_set != application_set
                    ):
                        raise ConflictingApplicationSetsError(
                            f'Application set {term.name} has a conflicting entry'
                        )

                if new_application_set:
                    new_application_set['name'] = term.name
                    self.applications.append(new_application_set)

            self.srx_policies.append((header, new_terms, filter_options))

    def _FixLargePolices(self, terms: list[policy.Term], address_family: str):
        """Loops over all terms finding terms exceeding SRXs policy limit.

        Args:
          terms: List of terms from a policy.
          address_family: Tuple containing address family versions.

        See the following URL for more information
        http://www.juniper.net/techpubs/en_US/junos12.1x44/topics/reference/
        general/address-address-sets-limitations.html
        """

        def Chunks(addresses: list[Union[nacaddr.IPv4, nacaddr.IPv6]]):
            """Splits a list of IP addresses into smaller lists based on byte size."""
            return_list = [[]]
            counter = 0
            index = 0
            for i in addresses:
                # Size is split in half due to the max size being a sum of src and dst.
                if counter > (self._ADDRESS_LENGTH_LIMIT / 2):
                    counter = 0
                    index += 1
                    return_list.append([])
                if i.version == 6:
                    counter += self._IPV6_SIZE
                else:
                    counter += 1
                return_list[index].append(i)
            return return_list

        expanded_terms = []
        for term in terms:
            if term.AddressesByteLength(self._AF_MAP[address_family]) > self._ADDRESS_LENGTH_LIMIT:
                logging.warning('LARGE TERM ENCOUNTERED')
                src_chunks = Chunks(term.source_address)
                counter = 0
                for chunk in src_chunks:
                    for ip in chunk:
                        ip.parent_token = f"src_{term.name}{counter!s}"
                    counter += 1
                dst_chunks = Chunks(term.destination_address)
                counter = 0
                for chunk in dst_chunks:
                    for ip in chunk:
                        ip.parent_token = f"dst_{term.name}{counter!s}"
                    counter += 1

                src_dst_products = itertools.product(src_chunks, dst_chunks)
                counter = 0
                for src_dst_list in src_dst_products:
                    new_term = copy.copy(term)
                    new_term.source_address = src_dst_list[0]
                    new_term.destination_address = src_dst_list[1]
                    new_term.name = f"{new_term.name}_{counter!s}"
                    expanded_terms.append(new_term)
                    counter += 1
            else:
                expanded_terms.append(term)
        if expanded_terms:
            del terms[:]
            terms.extend(expanded_terms)

    def _BuildPort(self, ports: list[tuple[int, int]]):
        """Transform specified ports into list and ranges.

        Args:
          ports: a policy terms list of ports

        Returns:
          port_list: list of ports and port ranges
        """
        port_list = []
        for i in ports:
            if i[0] == i[1]:
                port_list.append(str(i[0]))
            else:
                port_list.append(f'{str(i[0])}-{str(i[1])}')
        return port_list

    def _GenerateAddresses(self, token: str, ips, fqdns):
        target = IndentList(self.INDENT)
        counter = 0
        ips = nacaddr.SortAddrList(ips)
        ips = nacaddr.CollapseAddrList(ips)
        for ip in ips:
            target.IndentAppend(4, f"address {token}_{counter!s} {ip!s};")
            counter += 1
        for fqdn in fqdns:
            target.IndentAppend(4, f'address {token}_{counter} {{')
            target.IndentAppend(5, f'dns-name {fqdn.fqdn};')
            target.IndentAppend(4, f'}}')
        return target

    def _GenerateAddressSets(self, group, ips, fqdns) -> IndentList:
        target = IndentList(self.INDENT)
        target.IndentAppend(4, f"address-set {group} {{")
        counter = 0
        for _ in nacaddr.CollapseAddrList(ips):
            target.IndentAppend(5, f"address {group}_{counter!s};")
            counter += 1
        for _ in fqdns:
            target.IndentAppend(5, f"address {group}_{counter!s};")
            counter += 1
        target.IndentAppend(4, '}')
        return target

    def _GenerateAddressBook(self) -> IndentList:
        """Creates address book."""
        target = IndentList(self.INDENT)

        # create address books if address-book-type set to global
        if self._GLOBAL_ADDR_BOOK in self.addr_book_type:
            target.IndentAppend(1, 'replace: address-book {')
            target.IndentAppend(2, 'global {')
            global_addressbook = addressbook.Addressbook()
            for zone, _, ips, fqdns in self.addressbook.Walk():
                global_addressbook.AddAddresses('global', ips)
                global_addressbook.AddFQDNs('global', fqdns, suffix=FQDNSUFFIX)
            if 'global' in global_addressbook.GetZoneNames():
                addrs = []
                addr_sets = []
                for _, group, ips, fqdns in global_addressbook.Walk('global'):
                    addrs.extend(self._GenerateAddresses(group, ips, fqdns))
                    addr_sets.extend(self._GenerateAddressSets(group, ips, fqdns))
                target.extend(addrs)
                target.extend(addr_sets)

            target.IndentAppend(2, '}')
            target.IndentAppend(1, '}')

        else:
            target.IndentAppend(1, 'zones {')
            zones = self.addressbook.GetZoneNames()
            for zone in zones:
                target.IndentAppend(2, f"security-zone {zone} {{")
                target.IndentAppend(3, 'replace: address-book {')
                for _, group, ips, fqdns in self.addressbook.Walk(zone):
                    target.extend(self._GenerateAddresses(group, ips, fqdns))
                    target.extend(self._GenerateAddressSets(group, ips, fqdns))
                target.IndentAppend(3, '}')
                target.IndentAppend(2, '}')
            target.IndentAppend(1, '}')

        return target

    def _GenerateApplications(self) -> IndentList:
        target = IndentList(self.INDENT)
        apps_set_list = IndentList(self.INDENT)
        target.append('replace: applications {')
        done_apps = []
        for app in sorted(self.applications, key=lambda x: x['name']):
            app_list = IndentList(self.INDENT)
            if app in done_apps:
                continue

            if app['protocol'] or app['sport'] or app['dport'] or app['icmp-type']:
                # generate ICMP statements
                if app['icmp-type']:
                    if app['timeout']:
                        timeout = app['timeout']
                    else:
                        timeout = 60
                    # SRX has a limit of 8 terms per application. To get around this,
                    # we use application sets with applications that contain the terms
                    # we need.
                    num_terms = len(app['protocol']) * len(app['icmp-type'])
                    if num_terms > ICMP_TERM_LIMIT:
                        target.IndentAppend(1, f"application-set {app['name']}-app {{")
                        for i in range(num_terms):
                            target.IndentAppend(
                                2, f"application {app['name']}{'-app%d' % (i + 1)};"
                            )
                        target.IndentAppend(1, '}')
                    else:
                        target.IndentAppend(1, f"application {app['name']}-app {{")

                    term_counter = 0
                    for i, code in enumerate(app['icmp-type']):
                        for proto in app['protocol']:
                            # if we have more than 8 (ICMP_TERM_LIMIT) terms, we use an app
                            # for each term.
                            if num_terms > ICMP_TERM_LIMIT:
                                target.IndentAppend(
                                    1,
                                    'application '
                                    + app['name']
                                    + '-app%d' % (term_counter + 1)
                                    + ' {',
                                )
                                target.IndentAppend(
                                    2,
                                    'term t1 protocol %s %s-type %s inactivity-timeout %d;'
                                    % (proto, proto, str(code), int(timeout)),
                                )
                                target.IndentAppend(1, '}')
                            else:
                                target.IndentAppend(
                                    2,
                                    'term t%d protocol %s %s-type %s inactivity-timeout %d;'
                                    % (i + 1, proto, proto, str(code), int(timeout)),
                                )
                            term_counter += 1
                    if num_terms < ICMP_TERM_LIMIT:
                        target.IndentAppend(1, '}')
                # generate non-ICMP statements
                else:
                    i = 1
                    apps_set_list.IndentAppend(1, f"application-set {app['name']}-app {{")

                    for proto in app['protocol'] or ['']:
                        for sport in app['sport'] or ['']:
                            for dport in app['dport'] or ['']:
                                chunks = []
                                if proto:
                                    # SRX does not like proto vrrp
                                    if proto == 'vrrp':
                                        proto = '112'
                                    chunks.append(f' protocol {proto}')
                                if sport:
                                    chunks.append(f' source-port {sport}')
                                if dport:
                                    chunks.append(f' destination-port {dport}')
                                if app['timeout']:
                                    chunks.append(f" inactivity-timeout {int(app['timeout'])}")
                                if chunks:
                                    apps_set_list.IndentAppend(
                                        2, f"application {app['name']}{'-app%d;' % i}"
                                    )
                                    app_list.IndentAppend(
                                        1, f"application {app['name']}{'-app%d {' % i}"
                                    )

                                    app_list.IndentAppend(2, f"{'term t%d' % i}{''.join(chunks)};")
                                    app_list.IndentAppend(1, '}')
                                    i += 1
                    apps_set_list.IndentAppend(1, '}')

                done_apps.append(app)
                if app_list:
                    target.extend(app_list)

        if len(done_apps) == 0:
            target.clear()
            target.append('delete: applications;')
            return target

        target.extend(apps_set_list)
        target.append('}\n')
        return target

    def __str__(self) -> str:
        """Render the output of the JuniperSRX policy into config."""
        target = IndentList(self.INDENT)
        target.append('security {')

        # ADDRESSBOOK
        target.extend(self._GenerateAddressBook())

        # POLICIES
        target.IndentAppend(1, '/*')
        target.extend(aclgenerator.AddRepositoryTags(self.INDENT * 1))
        target.IndentAppend(1, '*/')

        target.IndentAppend(1, 'replace: policies {')

        for header, terms, filter_options in self.srx_policies:
            if self._NOVERBOSE not in filter_options[4:]:
                target.IndentAppend(2, '/*')
                target.extend(
                    [
                        self.INDENT * 2 + line
                        for line in aclgenerator.WrapWords(
                            header.comment, self._MAX_HEADER_COMMENT_LENGTH
                        )
                    ]
                )
                target.IndentAppend(2, '*/')

            # ZONE DIRECTION
            if filter_options[1] == 'all' and filter_options[3] == 'all':
                target.IndentAppend(2, 'global {')
            else:
                target.IndentAppend(
                    2, f"from-zone {filter_options[1]} to-zone {filter_options[3]} {{"
                )

            # GROUPS
            if header.apply_groups:
                target.IndentAppend(3, JunipersrxList('apply-groups', header.apply_groups))
            # GROUPS EXCEPT
            if header.apply_groups_except:
                target.IndentAppend(
                    3, JunipersrxList('apply-groups-except', header.apply_groups_except)
                )
            for term in terms:
                str_result = str(term)
                if str_result:
                    target.append(str_result)
            target.IndentAppend(2, '}')
        target.IndentAppend(1, '}')
        target.append('}')

        # APPLICATIONS
        target.extend(self._GenerateApplications())

        return '\n'.join(target)



# Copyright 2020-2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
"""Generic Google Cloud Platform multi-product generator.

Base class for GCP firewalling products.
"""

import json
import re

from aerleon.lib import aclgenerator
from aerleon.lib.policy import Term


class Error(aclgenerator.Error):
    """Generic error class."""


class TermError(Error):
    """Raised when a term is not valid."""


class HeaderError(Error):
    """Raised when a header is not valid."""


class UnsupportedFilterTypeError(Error):
    """Raised when an unsupported filter type is specified."""


class Term(aclgenerator.Term):
    """A Term object."""

    # Protocols allowed by name from:
    # https://cloud.google.com/vpc/docs/firewalls#protocols_and_ports
    # 'all' is needed for the dedault deny, it should not be used in a pol file.
    _ALLOW_PROTO_NAME = frozenset(['tcp', 'udp', 'icmp', 'esp', 'ah', 'ipip', 'sctp', 'all'])

    def _GetPorts(self) -> list[str]:
        """Return a port or port range in string format."""
        ports = []
        for start, end in self.term.destination_port:
            if start == end:
                ports.append(str(start))
            else:
                ports.append('%d-%d' % (start, end))
        return ports

    def _GetLoggingSetting(self) -> bool:
        """Return true if a term indicates that logging is desired."""
        # Supported values in GCP are '', 'true', and 'True'.
        settings = [str(x) for x in self.term.logging]
        if any(value in settings for value in ['true', 'True']):
            return True
        return False


class GCP(aclgenerator.ACLGenerator):
    """A GCP object."""

    policies = []
    _GOOD_DIRECTION = ['INGRESS', 'EGRESS']

    def __str__(self) -> str:
        """Return the JSON blob for a GCP object."""
        out = '%s\n\n' % (
            json.dumps(self.policies, indent=2, separators=(',', ': '), sort_keys=True)
        )
        return out


def IsDefaultDeny(term: Term) -> bool:
    """Return true if a term is a default deny without IPs, ports, etc."""
    skip_attrs = [
        'flattened',
        'flattened_addr',
        'flattened_saddr',
        'flattened_daddr',
        'action',
        'comment',
        'name',
        'logging',
    ]
    if 'deny' not in term.action:
        return False
    # This lc will look through all methods and attributes of the object.
    # It returns only the attributes that need to be looked at to determine if
    # this is a default deny.
    for i in [
        a
        for a in dir(term)
        if not a.startswith('__') and a.islower() and not callable(getattr(term, a))
    ]:
        if i in skip_attrs:
            continue
        v = getattr(term, i)
        if isinstance(v, str) and v:
            return False
        if isinstance(v, list) and v:
            return False

    return True


def IsProjectIDValid(project: str) -> bool:
    """Return true if a project ID is valid.

    https://cloud.google.com/resource-manager/reference/rest/v1/projects

    "It must be 6 to 30 lowercase letters, digits, or hyphens. It must start with
    a letter. Trailing hyphens are prohibited."

    Args:
      project: A string.

    Returns:
      bool: True if a project ID matches the pattern and length requirements.
    """
    if len(project) < 6 or len(project) > 30:
        return False
    return bool(re.match('^[a-z][a-z0-9\\-]*[a-z0-9]$', project))


def IsVPCNameValid(vpc: str) -> bool:
    """Return true if a VPC name is valid.

    https://cloud.google.com/compute/docs/reference/rest/v1/networks

    "The first character must be a lowercase letter, and all following characters
    (except for the last character) must be a dash, lowercase letter, or digit.
    The last character must be a lowercase letter or digit."

    Args:
      vpc: A string.

    Returns:
      bool: True if a VPC name matches the pattern and length requirements.
    """
    if len(vpc) < 1 or len(vpc) > 63:
        return False
    return bool(re.match('^[a-z]$|^[a-z][a-z0-9-]*[a-z0-9]$', vpc))


def TruncateString(raw_string: str, max_length: int) -> str:
    """Returns truncated raw_string based on max length.

    Args:
      raw_string: String to be truncated.
      max_length: max length of string.

    Returns:
      string: The truncated string.
    """
    if len(raw_string) > max_length:
        return raw_string[:max_length]
    return raw_string


def GetIpv6TermName(term_name: str) -> str:
    """Returns the equivalent term name for IPv6 terms.

    Args:
      term_name: A string.

    Returns:
      string: The IPv6 requivalent term name.
    """

    return f'{term_name}-v6'



# Copyright 2017 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Palo Alto Firewall generator."""

import collections
import copy
import re
import xml.etree.ElementTree as etree
from typing import Optional, Union
from xml.dom import minidom

from absl import logging

from aerleon.lib import aclgenerator, addressbook, nacaddr, policy
from aerleon.lib.policy import Policy, Term


class Error(aclgenerator.Error):
    """generic error class."""


class UnsupportedFilterError(Error):
    pass


class UnsupportedHeaderError(Error):
    pass


class PaloAltoFWDuplicateTermError(Error):
    pass


class PaloAltoFWUnsupportedProtocolError(Error):
    pass


class PaloAltoFWVerbatimError(Error):
    pass


class PaloAltoFWOptionError(Error):
    pass


class PaloAltoFWDuplicateServiceError(Error):
    pass


class PaloAltoFWNameTooLongError(Error):
    pass


class PaloAltoFWBadIcmpTypeError(Error):
    pass


class ServiceMap:
    """Manages service names across a single policy instance."""

    def __init__(self) -> None:
        self.entries = {}

    def get_service_name(
        self,
        term_name: str,
        src_ports: tuple[str],
        ports: Union[tuple[str, str], tuple[str]],
        protocol: str,
        prefix: Optional[str] = None,
    ) -> str:
        """Returns service name based on the provided ports and protocol."""
        if (src_ports, ports, protocol) in self.entries:
            return self.entries[(src_ports, ports, protocol)]["name"]

        if prefix is None:
            prefix = "service-"
        service_name = f"{prefix}{term_name}-{protocol}"

        if len(service_name) > 63:
            raise PaloAltoFWNameTooLongError(
                f"Service name must be 63 characters max: {service_name}"
            )

        for _, service in self.entries.items():
            if service["name"] == service_name:
                raise PaloAltoFWDuplicateServiceError(
                    "You have a duplicate service. A service named %s already exists."
                    % service_name
                )

        self.entries[(src_ports, ports, protocol)] = {"name": service_name}
        return service_name


class Rule:
    """Extend the Term() class for PaloAlto Firewall Rules."""

    def __init__(self, from_zone: str, to_zone: str, term: Term, service_map: ServiceMap) -> None:
        # Palo Alto Firewall rule keys
        MAX_ZONE_LENGTH = 31

        if not from_zone or not to_zone:
            raise PaloAltoFWOptionError("Source or destination zone is empty.")
        if len(from_zone) > MAX_ZONE_LENGTH:
            x = "Source zone must be %d characters max: %s" % (MAX_ZONE_LENGTH, from_zone)
            raise PaloAltoFWNameTooLongError(x)
        if len(to_zone) > MAX_ZONE_LENGTH:
            x = "Destination zone must be %d characters max: %s" % (MAX_ZONE_LENGTH, to_zone)
            raise PaloAltoFWNameTooLongError(x)

        self.options = []

        while term is not None:
            x, term = self.TermToOptions(from_zone, to_zone, term, service_map)
            self.options.append(x)

    @staticmethod
    def TermToOptions(
        from_zone: str, to_zone: str, term: Term, service_map: ServiceMap
    ) -> tuple[dict[str, Union[list[str], str]], Optional[Term]]:
        """Convert term to Palo Alto security rule options."""
        options = {}
        options["from_zone"] = [from_zone]
        options["to_zone"] = [to_zone]
        options["description"] = []
        options["source"] = []
        options["destination"] = []
        options["application"] = []
        options["service"] = []
        options["logging"] = []

        ACTIONS = {
            "accept": "allow",
            "deny": "deny",
            "reject": "reset-client",
            "reject-with-tcp-rst": "reset-client",
        }

        new_term = None

        def pan_ports(ports):
            x = []
            for tup in ports:
                if len(tup) > 1 and tup[0] != tup[1]:
                    x.append(f"{tup[0]!s}-{tup[1]!s}")
                else:
                    x.append(str(tup[0]))

            return tuple(x)

        # COMMENT
        if term.comment:
            options["description"] = term.comment

        # LOGGING
        if term.logging:
            for item in term.logging:
                if item.value in ["disable"]:
                    options["logging"] = ["disable"]
                    break
                elif item.value in ["log-both"]:
                    options["logging"].append("log-start")
                    options["logging"].append("log-end")
                elif item.value in ["True", "true", "syslog", "local"]:
                    options["logging"].append("log-end")

        # SOURCE-ADDRESS
        if term.source_address:
            saddr_check = set()
            for saddr in term.source_address:
                saddr_check.add(saddr.parent_token)
            saddr_check = sorted(saddr_check)
            for addr in saddr_check:
                options["source"].append(str(addr))
        # missing source handled during XML document generation

        # DESTINATION-ADDRESS
        if term.destination_address:
            daddr_check = set()
            for daddr in term.destination_address:
                daddr_check.add(daddr.parent_token)
            daddr_check = sorted(daddr_check)
            for addr in daddr_check:
                options["destination"].append(str(addr))
        # missing destination handled during XML document generation

        # ACTION
        if term.action:
            options["action"] = ACTIONS[term.action[0]]

        if term.option:
            options["option"] = term.option

        if term.pan_application:
            for pan_app in term.pan_application:
                options["application"].append(pan_app)

        if term.source_port or term.destination_port:
            src_ports = pan_ports(term.source_port)
            if term.destination_port:
                ports = pan_ports(term.destination_port)
            else:
                ports = pan_ports([("0", "65535")])

            # check to see if this service already exists
            for p in sorted(term.protocol):
                service_name = service_map.get_service_name(term.name, src_ports, ports, p)
                if service_name not in options["service"]:
                    options["service"].append(service_name)

        elif "tcp" in term.protocol or "udp" in term.protocol:
            services = {"tcp", "udp"} & set(term.protocol)
            others = set(term.protocol) - services
            if others:
                logging.info(
                    "INFO: Term %s in policy %s>%s contains port-less %s "
                    "with non-port protocol(s). Moving %s to a new term.",
                    term.name,
                    from_zone,
                    to_zone,
                    ', '.join(list(services)),
                    ', '.join(list(others)),
                )
                new_term = copy.deepcopy(term)
                new_term.protocol = list(others)
                term.protocol = list(services)
                options["application"] = []
            for p in sorted(term.protocol):
                ports = pan_ports([("0", "65535")])
                # use prefix "" to avoid service name clash with term named "any"
                service_name = service_map.get_service_name("any", (), ports, p, "")
                if service_name not in options["service"]:
                    options["service"].append(service_name)

        if term.protocol:
            # Add certain protocol names as application in the application list
            # if missing.
            for proto_name in term.protocol:
                if (
                    proto_name in ["igmp", "sctp", "gre"]
                    and proto_name not in options["application"]
                ):
                    options["application"].append(proto_name)
                elif proto_name in ("ah", "esp"):
                    ipsec_app_proto = f"ipsec-{proto_name}"
                    if ipsec_app_proto not in options["application"]:
                        options["application"].append(ipsec_app_proto)

        return options, new_term


class PaloAltoFW(aclgenerator.ACLGenerator):
    """PaloAltoFW rendering class."""

    _PLATFORM = "paloalto"
    SUFFIX = ".xml"
    _SUPPORTED_AF = {"inet", "inet6", "mixed"}
    _AF_MAP = {"inet": (4,), "inet6": (6,), "mixed": (4, 6)}
    _TERM_MAX_LENGTH = 63
    _APPLICATION_NAME_MAX_LENGTH = 31
    _TERM_PREFIX_LENGTH = 24
    _SUPPORTED_PROTO_NAMES = [
        "tcp",
        "udp",
        "icmp",
        "icmpv6",
        "sctp",
        "igmp",
        "gre",
        "ah",
        "esp",
    ]
    _MAX_RULE_DESCRIPTION_LENGTH = 1024
    _MAX_TAG_COMMENTS_LENGTH = 1023
    _TAG_NAME_FORMAT = "{from_zone}_{to_zone}_policy-comment-{num}"
    _MAX_RULE_SRC_DST_MEMBERS = 65535

    _ABBREVIATION_TABLE = [
        # Service abbreviations first.
        ("experiment", "EXP"),
        ("wifi-radius", "W-R"),
        ("customer", "CUST"),
        ("server", "SRV"),
        # Next, common routing terms
        ("global", "GBL"),
        ("google", "GOOG"),
        ("service", "SVC"),
        ("router", "RTR"),
        ("internal", "INT"),
        ("external", "EXT"),
        ("transit", "TRNS"),
        ("management", "MGMT"),
        # State info
        ("established", "EST"),
        ("unreachable", "UNR"),
        ("fragment", "FRAG"),
        ("accept", "ACC"),
        ("discard", "DISC"),
        ("reject", "REJ"),
        ("replies", "RPLS"),
        ("reply", "RPL"),
        ("request", "REQ"),
        # ICMP types specific
        ("inverse", "INV"),
        ("neighbor", "NBR"),
        ("discovery", "DSCVR"),
        ("advertisement", "ADV"),
        ("solicitation", "SOL"),
        ("multicast", "MCAST"),
        ("certification", "CERT"),
        ("listener", "LSNR"),
        ("address", "ADDR"),
    ]

    INDENT = "  "

    def __init__(self, pol: Policy, exp_info: int) -> None:
        self.pafw_policies = []
        self.addressbook = addressbook.Addressbook()
        self.applications = []
        self.application_refs = {}
        self.application_groups = []
        self.pan_applications = []
        self.ports = []
        self.from_zone = ""
        self.to_zone = ""
        self.policy_name = ""
        self.config = None
        self.service_map = ServiceMap()
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens = {
            "action",
            "comment",
            "destination_address",
            "destination_address_exclude",
            "destination_port",
            "expiration",
            "icmp_type",
            "logging",
            "name",
            "option",
            "owner",
            "platform",
            "platform_exclude",
            "protocol",
            "source_address",
            "source_address_exclude",
            "source_port",
            "stateless_reply",
            "timeout",
            "pan_application",
            "translated",
        }

        supported_sub_tokens.update(
            {
                "action": {"accept", "deny", "reject", "reject-with-tcp-rst"},
                "option": {"established", "tcp-established"},
            }
        )
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        """Transform a policy object into a PaloAltoFW object.

        Args:
          pol: policy.Policy object
          exp_info: print a info message when a term is set to expire in that many
            weeks

        Raises:
          UnsupportedFilterError: An unsupported filter was specified
          UnsupportedHeaderError: A header option exists that is not
          understood/usable
          PaloAltoFWDuplicateTermError: Two terms were found with same name in
          same filter
          PaloAltoFWBadIcmpTypeError: The referenced ICMP type is not supported
          by the policy term.
          PaloAltoFWUnsupportedProtocolError: The term contains unsupporter protocol
          name.
        """
        first_addr_obj = None
        address_book_dup_check = set()

        for header, terms in pol.filters:
            # The filter_options is a list of options from header, e.g.
            # ['from-zone', 'internal', 'to-zone', 'external']
            filter_options = header.FilterOptions(self._PLATFORM)

            if (
                len(filter_options) < 4
                or filter_options[0] != "from-zone"
                or filter_options[2] != "to-zone"
            ):
                raise UnsupportedFilterError(
                    "Palo Alto Firewall filter arguments must specify from-zone and " "to-zone."
                )

            self.from_zone = filter_options[1]
            self.to_zone = filter_options[3]

            # The filter_type values are either inet, inet6, or mixed. Later, the
            # code analyzes source and destination IP addresses and determines whether
            # it is an appropriate type for the filter_type value.
            if len(filter_options) > 4:
                filter_type = filter_options[4]
            else:
                filter_type = "inet"

            if filter_type not in self._SUPPORTED_AF:
                raise UnsupportedHeaderError(
                    'Palo Alto Firewall Generator invalid address family option: "%s"'
                    '; expect {%s}' % (filter_type, '|'.join(self._SUPPORTED_AF))
                )

            valid_addr_obj = ["addr-obj", "no-addr-obj"]
            if len(filter_options) > 5 and filter_options[5] not in valid_addr_obj:
                raise UnsupportedHeaderError(
                    'Palo Alto Firewall Generator invalid address objects option: "%s"'
                    '; expect {%s}' % (filter_options[5], '|'.join(valid_addr_obj))
                )
            no_addr_obj = (
                True if (len(filter_options) > 5 and filter_options[5] == "no-addr-obj") else False
            )

            unique_term_prefixes = (
                True
                if (len(filter_options) > 6 and filter_options[6] == "unique-term-prefixes")
                else False
            )

            if first_addr_obj is None:
                first_addr_obj = no_addr_obj
            if first_addr_obj != no_addr_obj:
                raise UnsupportedHeaderError(
                    "Cannot mix addr-obj and no-addr-obj header option in " "a single policy file"
                )

            term_dup_check = set()
            new_terms = []

            for term in terms:
                if term.stateless_reply:
                    logging.warning(
                        "WARNING: Term %s in policy %s>%s is a stateless reply "
                        "term and will not be rendered.",
                        term.name,
                        self.from_zone,
                        self.to_zone,
                    )
                    continue
                if "established" in term.option:
                    logging.warning(
                        "WARNING: Term %s in policy %s>%s is a established "
                        "term and will not be rendered.",
                        term.name,
                        self.from_zone,
                        self.to_zone,
                    )
                    continue
                if "tcp-established" in term.option:
                    logging.warning(
                        "WARNING: Term %s in policy %s>%s is a tcp-established "
                        "term and will not be rendered.",
                        term.name,
                        self.from_zone,
                        self.to_zone,
                    )
                    continue

                if unique_term_prefixes:
                    # Prefix hash of from_zone and to_zone to the term name to get mostly
                    # unique names across different policies.
                    # This is not crytopgraphically guaranteed.
                    # This modified term name should not exceed _TERM_MAX_LENGTH.
                    term.name = (
                        self.HexDigest(self.from_zone + self.to_zone, self._TERM_PREFIX_LENGTH)
                        + "-"
                        + term.name
                    )
                term.name = self.FixTermLength(term.name)
                if term.name in term_dup_check:
                    raise PaloAltoFWDuplicateTermError(f"You have a duplicate term: {term.name}")
                term_dup_check.add(term.name)

                services = {"tcp", "udp"} & set(term.protocol)
                others = set(term.protocol) - services
                if others and term.pan_application:
                    raise UnsupportedFilterError(
                        "Term %s contains non tcp, udp protocols with pan-application: %s: %s"
                        "\npan-application can only be used with protocols tcp, udp"
                        % (term.name, ', '.join(term.pan_application), ', '.join(term.protocol))
                    )

                for i in term.source_address_exclude:
                    term.source_address = nacaddr.RemoveAddressFromList(term.source_address, i)
                for i in term.destination_address_exclude:
                    term.destination_address = nacaddr.RemoveAddressFromList(
                        term.destination_address, i
                    )

                # Count the number of occurencies of a particular version of the
                # address family, i.e. v4/v6 in source and destination IP addresses.
                afc = {
                    4: {"src": 0, "dst": 0},
                    6: {"src": 0, "dst": 0},
                }
                # Determine the address families in the source and destination
                # addresses references in the term. Next, determine IPv4 and IPv6
                # traffic flow patterns.
                exclude_address_family = []
                flows = []
                src_any = False
                dst_any = False
                if not term.source_address:
                    src_any = True
                if not term.destination_address:
                    dst_any = True
                for addr in term.source_address:
                    afc[addr.version]["src"] += 1
                for addr in term.destination_address:
                    afc[addr.version]["dst"] += 1
                for v in [4, 6]:
                    if src_any and dst_any:
                        flows.append("ip%d-ip%d" % (v, v))
                        continue
                    if (afc[v]["src"] == 0 and not src_any) and (
                        afc[v]["dst"] == 0 and not dst_any
                    ):
                        continue
                    if (afc[v]["src"] > 0 or src_any) and (afc[v]["dst"] > 0 or dst_any):
                        flows.append("ip%d-ip%d" % (v, v))
                        continue
                    if (afc[v]["src"] > 0 or src_any) and afc[v]["dst"] == 0:
                        flows.append("ip%d-src-only" % v)
                        flows.append("ip%d-only" % v)
                        continue
                    if afc[v]["src"] == 0 and (afc[v]["dst"] > 0 or dst_any):
                        flows.append("ip%d-dst-only" % v)
                        flows.append("ip%d-only" % v)

                if filter_type == "inet":
                    if "icmpv6" in term.protocol:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s references ICMPv6 protocol, "
                            "term will not be rendered.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                        )
                        continue
                    if "ip4-ip4" not in flows:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s has one or more invalid "
                            "src-dest combinations %s, term will not be rendered.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                            flows,
                        )
                        continue
                    # exclude IPv6 addresses
                    exclude_address_family.append(6)
                elif filter_type == "inet6":
                    if "icmp" in term.protocol:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s references ICMP protocol, "
                            "term and will not be rendered.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                        )
                        continue
                    if "ip6-ip6" not in flows:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s has one or more invalid "
                            "src-dest combinations %s, term will not be rendered.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                            flows,
                        )
                        continue
                    exclude_address_family.append(4)
                elif filter_type == "mixed":
                    if "ip4-ip4" in flows and "ip6-ip6" not in flows:
                        exclude_address_family.append(6)
                        pass
                    elif "ip6-ip6" in flows and "ip4-ip4" not in flows:
                        exclude_address_family.append(4)
                        pass
                    elif "ip4-ip4" in flows and "ip6-ip6" in flows:
                        pass
                    elif "ip4-only" in flows and "ip6-only" in flows:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s has source and destinations "
                            "of different address families %s, term will not be "
                            "rendered.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                            filter(lambda p: re.search(p, "(src|dst)-only"), flows),
                        )
                        continue
                    else:
                        logging.warning(
                            "WARNING: Term %s in policy %s>%s has invalid src-dest "
                            "combinations %s, the term will be rendered without them.",
                            term.name,
                            self.from_zone,
                            self.to_zone,
                            filter(lambda p: re.search(p, "(src|dst)-only"), flows),
                        )
                        if "ip4-ip4" in flows:
                            exclude_address_family.append(6)
                        else:
                            exclude_address_family.append(4)

                # Substitute large IPv6 ranges (/1, /2) with equivalent subnets.
                # Do this separately from address book building, or during policy
                # translation to account for both address-objects and no-address-objects
                if term.source_address:
                    saddr_split = []
                    for saddr in term.source_address:
                        if saddr.version == 6 and 0 < saddr.prefixlen < 3:
                            for saddr2 in saddr.subnets(new_prefix=3):
                                saddr2.parent_token = saddr.parent_token
                                saddr_split.append(saddr2)
                        else:
                            saddr_split.append(saddr)
                    term.source_address = saddr_split
                    source_address = [
                        addr
                        for addr in term.source_address
                        if addr.version not in exclude_address_family
                    ]
                    if source_address:
                        self.addressbook.AddAddresses(
                            '',
                            source_address,
                        )
                if term.destination_address:
                    daddr_split = []
                    for daddr in term.destination_address:
                        if daddr.version == 6 and 0 < daddr.prefixlen < 3:
                            for daddr2 in daddr.subnets(new_prefix=3):
                                daddr2.parent_token = daddr.parent_token
                                daddr_split.append(daddr2)
                        else:
                            daddr_split.append(daddr)
                    term.destination_address = daddr_split
                    destination_address = [
                        addr
                        for addr in term.destination_address
                        if addr.version not in exclude_address_family
                    ]
                    if destination_address:
                        self.addressbook.AddAddresses(
                            '',
                            destination_address,
                        )

                # Handle ICMP/ICMPv6 terms.
                if term.icmp_type and (
                    "icmp" not in term.protocol and "icmpv6" not in term.protocol
                ):
                    raise UnsupportedFilterError(
                        "Palo Alto Firewall filter must have ICMP or ICMPv6 protocol "
                        + "specified when using icmp_type keyword"
                    )

                for icmp_version in ["icmp", "icmpv6"]:
                    if "icmp" not in term.protocol and "icmpv6" not in term.protocol:
                        # the protocol is not ICMP or ICMPv6
                        break
                    if icmp_version not in term.protocol:
                        # skip if this icmp_version isn't in the term protocol.
                        continue
                    if icmp_version == "icmp" and "ip4-ip4" not in flows:
                        # skip if there is no ip4 to ipv4 communication
                        continue
                    if icmp_version == "icmpv6" and "ip6-ip6" not in flows:
                        # skip if there is no ip4 to ipv4 communication
                        continue
                    if icmp_version == "icmp":
                        if filter_type == "inet6":
                            continue
                        if not term.icmp_type:
                            term.pan_application.append("icmp")
                            continue
                        icmp_type_keyword = "ident-by-icmp-type"
                        # The risk level 4 is the default PANOS' risk level for ICMP.
                        risk_level = 4
                    else:
                        if filter_type == "inet":
                            continue
                        if not term.icmp_type:
                            term.pan_application.append("ipv6-icmp")
                            continue
                        icmp_type_keyword = "ident-by-icmp6-type"
                        # The risk level 2 is the default PANOS' risk level for ICMPv6.
                        risk_level = 2
                    # The term contains ICMP types
                    for term_icmp_type_name in term.icmp_type:
                        if icmp_version == "icmp":
                            icmp_app_name = f"icmp-{term_icmp_type_name}"
                            # This is to abbreviate the Application name where possible.
                            # The limit is defined by _APPLICATION_NAME_MAX_LENGTH = 31.
                            if len(icmp_app_name) > self._APPLICATION_NAME_MAX_LENGTH:
                                icmp_app_name = self.FixTermLength(
                                    icmp_app_name, True, True, self._APPLICATION_NAME_MAX_LENGTH
                                )
                            if term_icmp_type_name not in policy.Term.ICMP_TYPE[4]:
                                raise PaloAltoFWBadIcmpTypeError(
                                    "term with bad icmp type: %s, icmp_type: %s"
                                    % (term.name, term_icmp_type_name)
                                )
                            term_icmp_type = policy.Term.ICMP_TYPE[4][term_icmp_type_name]
                        else:
                            icmp_app_name = f"icmp6-{term_icmp_type_name}"
                            # This is to abbreviate the Application name where possible.
                            # The limit is defined by _APPLICATION_NAME_MAX_LENGTH = 31.
                            if len(icmp_app_name) > self._APPLICATION_NAME_MAX_LENGTH:
                                icmp_app_name = self.FixTermLength(
                                    icmp_app_name, True, True, self._APPLICATION_NAME_MAX_LENGTH
                                )
                            if term_icmp_type_name not in policy.Term.ICMP_TYPE[6]:
                                raise PaloAltoFWBadIcmpTypeError(
                                    "term with bad icmp type: %s, icmp_type: %s"
                                    % (term.name, term_icmp_type_name)
                                )
                            term_icmp_type = policy.Term.ICMP_TYPE[6][term_icmp_type_name]
                        if icmp_app_name not in self.application_refs:
                            # the custom icmp application doesn't already exist
                            app_entry = {
                                "category": "networking",
                                "subcategory": "ip-protocol",
                                "technology": "network-protocol",
                                "description": icmp_app_name,
                                "default": {
                                    icmp_type_keyword: "%d" % term_icmp_type,
                                },
                                "risk": "%d" % risk_level,
                            }
                            self.application_refs[icmp_app_name] = app_entry
                            self.applications.append(icmp_app_name)

                        # always add the ICMP application to the term, it either already
                        # existed due to a previous policy, or it was created in the
                        # previous loop.
                        if icmp_app_name not in term.pan_application:
                            term.pan_application.append(icmp_app_name)

                # Filter out unsupported protocols
                for proto_name in term.protocol:
                    if proto_name in self._SUPPORTED_PROTO_NAMES:
                        continue
                    raise PaloAltoFWUnsupportedProtocolError(
                        f"protocol {proto_name} is not supported"
                    )

                if term.icmp_type:
                    if set(term.protocol) == {'icmp', 'icmpv6'}:
                        raise UnsupportedFilterError(
                            '%s %s'
                            % (
                                'icmp-type specified for both icmp and icmpv6 protocols'
                                ' in a single term:',
                                term.name,
                            )
                        )
                    if term.protocol != ['icmp'] and term.protocol != ['icmpv6']:
                        raise UnsupportedFilterError(
                            '%s %s'
                            % ('icmp-type specified for non-icmp protocols in term:', term.name)
                        )

                new_terms.append(term)

            # Create a ruleset. It contains the rules for the terms defined under
            # a single header on a particular platform.
            ruleset = {}

            for term in new_terms:
                current_rule = Rule(self.from_zone, self.to_zone, term, self.service_map)
                if len(current_rule.options) > 1:
                    for i, v in enumerate(current_rule.options):
                        name = "%s-%d" % (term.name, i + 1)
                        name = self.FixTermLength(name)
                        ruleset[name] = v
                else:
                    ruleset[term.name] = current_rule.options[0]

            self.pafw_policies.append((header, ruleset, filter_options))

    def _SortAddressBookNumCheck(self, item: str) -> tuple[str, int]:
        """Used to give a natural order to the list of acl entries.

        Args:
          item: string of the address book entry name

        Returns:
          returns the characters and number
        """
        item_list = item.split("_")
        num = item_list.pop(-1)
        if isinstance(item_list[-1], int):
            set_number = item_list.pop(-1)
            num = int(set_number) * 1000 + int(num)
        alpha = "_".join(item_list)
        if num:
            return (alpha, int(num))
        return (alpha, 0)

    def _BuildPort(self, ports):
        """Transform specified ports into list and ranges.

        Args:
          ports: a policy terms list of ports

        Returns:
          port_list: list of ports and port ranges
        """
        port_list = []
        for i in ports:
            if i[0] == i[1]:
                port_list.append(str(i[0]))
            else:
                port_list.append(f"{str(i[0])}-{str(i[1])}")
        return port_list

    def __str__(self) -> str:
        """Render the output of the PaloAltoFirewall policy into config."""

        # IPv4 addresses are normalized into the policy as IPv6 addresses
        # using ::<ipv4-address>.  The 0.0.0.0-255.255.255.255 range is
        # equivalent to ::0/96 which will only match IPv4 addresses; when
        # negated it will match only IPv6 addresses.
        # Used for address families inet and inet6 when source and
        # destination address are not specified (any any).
        ANY_IPV4_RANGE = "0.0.0.0-255.255.255.255"
        add_any_ipv4 = False

        # INITAL CONFIG
        config = etree.Element("config", {"urldb": "paloaltonetworks", "version": "8.1.0"})
        devices = etree.SubElement(config, "devices")
        device_entry = etree.SubElement(devices, "entry", {"name": "localhost.localdomain"})
        vsys = etree.SubElement(device_entry, "vsys")
        vsys_entry = etree.SubElement(vsys, "entry", {"name": "vsys1"})

        # APPLICATION
        app_entries = etree.Element("application")
        for app_name in self.applications:
            if app_name not in self.application_refs:
                # this is not a custom application.
                continue
            app = self.application_refs[app_name]
            app_entry = etree.SubElement(app_entries, "entry", {"name": app_name})
            for k in self.application_refs[app_name]:
                if isinstance(app[k], (str)):
                    etree.SubElement(app_entry, k).text = app[k]
                elif isinstance(app[k], (dict)):
                    if k == "default":
                        default_props = etree.SubElement(app_entry, "default")
                    else:
                        continue
                    for prop in app[k]:
                        if k == "default" and prop in [
                            "ident-by-icmp-type",
                            "ident-by-icmp6-type",
                        ]:
                            icmp_type_props = etree.SubElement(default_props, prop)
                            etree.SubElement(icmp_type_props, "type").text = app[k][prop]
                        else:
                            pass
        vsys_entry.append(app_entries)

        # APPLICATION GROUPS
        etree.SubElement(vsys_entry, "application-group")

        # SERVICES
        vsys_entry.append(etree.Comment(" Services "))
        service = etree.SubElement(vsys_entry, "service")
        for k, v in self.service_map.entries.items():
            entry = etree.SubElement(service, "entry", {"name": v["name"]})
            proto0 = etree.SubElement(entry, "protocol")
            proto = etree.SubElement(proto0, k[2])
            # destination port
            port = etree.SubElement(proto, "port")
            tup = str(k[1])[1:-1]
            if tup[-1] == ",":
                tup = tup[:-1]
            port.text = tup.replace("'", "").replace(", ", ",")
            # source port
            if len(k[0]):
                sport = etree.SubElement(proto, "source-port")
                tup = str(k[0])[1:-1]
                if tup[-1] == ",":
                    tup = tup[:-1]
                sport.text = tup.replace("'", "").replace(", ", ",")

        # RULES
        vsys_entry.append(etree.Comment(" Rules "))
        rulebase = etree.SubElement(vsys_entry, "rulebase")
        security = etree.SubElement(rulebase, "security")
        rules = etree.SubElement(security, "rules")
        tag = etree.Element("tag")

        tag_num = 0

        # pytype: disable=key-error
        # pylint: disable=unused-variable
        for header, pa_rules, filter_options in self.pafw_policies:
            tag_name = None
            if header.comment:
                comment = " ".join(header.comment).strip()
                if comment:
                    tag_num += 1
                    # max tag len 127, max zone len 31
                    tag_name = self._TAG_NAME_FORMAT.format(
                        from_zone=filter_options[1], to_zone=filter_options[3], num=tag_num
                    )
                    tag_entry = etree.SubElement(tag, "entry", {"name": tag_name})
                    comments = etree.SubElement(tag_entry, "comments")
                    if len(comment) > self._MAX_TAG_COMMENTS_LENGTH:
                        logging.warning(
                            "WARNING: tag %s comments exceeds maximum " "length %d, truncated.",
                            tag_name,
                            self._MAX_TAG_COMMENTS_LENGTH,
                        )
                    comments.text = comment[: self._MAX_TAG_COMMENTS_LENGTH]

            no_addr_obj = (
                True if (len(filter_options) > 5 and filter_options[5] == "no-addr-obj") else False
            )

            for name, options in pa_rules.items():
                entry = etree.SubElement(rules, "entry", {"name": name})
                if options["description"]:
                    descr = etree.SubElement(entry, "description")
                    x = " ".join(options["description"])
                    if len(x) > self._MAX_RULE_DESCRIPTION_LENGTH:
                        logging.warning(
                            "WARNING: rule %s description exceeds maximum "
                            "length %d, truncated.",
                            name,
                            self._MAX_RULE_DESCRIPTION_LENGTH,
                        )
                    descr.text = x[: self._MAX_RULE_DESCRIPTION_LENGTH]

                to = etree.SubElement(entry, "to")
                for x in options["to_zone"]:
                    member = etree.SubElement(to, "member")
                    member.text = x

                from_ = etree.SubElement(entry, "from")
                for x in options["from_zone"]:
                    member = etree.SubElement(from_, "member")
                    member.text = x

                af = filter_options[4] if len(filter_options) > 4 else "inet"

                max_src_dst = 0
                source = etree.SubElement(entry, "source")
                if not options["source"]:
                    member = etree.SubElement(source, "member")
                    if not options["destination"] and af != "mixed":
                        # only inet and inet6 use the any-ipv4 object
                        member.text = "any-ipv4"
                        add_any_ipv4 = True
                    else:
                        member.text = "any"
                else:
                    for x in options["source"]:
                        if no_addr_obj:
                            for ip in self.addressbook.GetAddress('', x):
                                member = etree.SubElement(source, "member")
                                member.text = str(ip)
                                max_src_dst += 1
                        else:
                            member = etree.SubElement(source, "member")
                            member.text = x
                            max_src_dst += 1

                if max_src_dst > self._MAX_RULE_SRC_DST_MEMBERS:
                    raise UnsupportedFilterError(
                        "term %s source members exceeds maximum of %d: %d"
                        % (name, self._MAX_RULE_SRC_DST_MEMBERS, max_src_dst)
                    )

                max_src_dst = 0
                dest = etree.SubElement(entry, "destination")
                if not options["destination"]:
                    member = etree.SubElement(dest, "member")
                    if options["source"]:
                        member.text = "any"
                    else:
                        if af != "mixed":
                            # only inet and inet6 use the any-ipv4 object
                            member.text = "any-ipv4"
                            if af == "inet6":
                                for x in ["negate-source", "negate-destination"]:
                                    negate = etree.SubElement(entry, x)
                                    negate.text = "yes"
                        else:
                            member.text = "any"
                else:
                    for x in options["destination"]:
                        if no_addr_obj:
                            for ip in self.addressbook.GetAddress('', x):
                                member = etree.SubElement(dest, "member")
                                member.text = str(ip)
                                max_src_dst += 1
                        else:
                            member = etree.SubElement(dest, "member")
                            member.text = x
                            max_src_dst += 1

                if max_src_dst > self._MAX_RULE_SRC_DST_MEMBERS:
                    raise UnsupportedFilterError(
                        "term %s destination members exceeds maximum of %d: %d"
                        % (name, self._MAX_RULE_SRC_DST_MEMBERS, max_src_dst)
                    )

                # service section of a policy rule.
                service = etree.SubElement(entry, "service")
                if not options["service"] and not options["application"]:
                    member = etree.SubElement(service, "member")
                    member.text = "any"
                elif not options["service"] and options["application"]:
                    # Adds custom applications.
                    member = etree.SubElement(service, "member")
                    member.text = "application-default"
                else:
                    # Adds services.
                    for x in options["service"]:
                        member = etree.SubElement(service, "member")
                        member.text = x

                # ACTION
                action = etree.SubElement(entry, "action")
                action.text = options["action"]

                # check whether the rule is interzone
                if list(set(options["from_zone"]).difference(options["to_zone"])):
                    type_ = etree.SubElement(entry, "rule-type")
                    type_.text = "interzone"
                elif not options["from_zone"] and not options["to_zone"]:
                    type_ = etree.SubElement(entry, "rule-type")
                    type_.text = "interzone"

                # APPLICATION
                app = etree.SubElement(entry, "application")
                if not options["application"]:
                    member = etree.SubElement(app, "member")
                    member.text = "any"
                else:
                    for x in sorted(options["application"]):
                        member = etree.SubElement(app, "member")
                        member.text = x

                if tag_name is not None:
                    rules_tag = etree.SubElement(entry, "tag")
                    member = etree.SubElement(rules_tag, "member")
                    member.text = tag_name

                # LOGGING
                if options["logging"]:
                    if "disable" in options["logging"]:
                        log = etree.SubElement(entry, "log-start")
                        log.text = "no"
                        log = etree.SubElement(entry, "log-end")
                        log.text = "no"
                    if "log-start" in options["logging"]:
                        log = etree.SubElement(entry, "log-start")
                        log.text = "yes"
                    if "log-end" in options["logging"]:
                        log = etree.SubElement(entry, "log-end")
                        log.text = "yes"

        # pytype: enable=key-error

        # ADDRESS
        vsys_entry.append(etree.Comment(" Address Groups "))
        addr_group = etree.SubElement(vsys_entry, "address-group")

        if not no_addr_obj:
            for _, token, ips, _ in self.addressbook.Walk(''):
                entry = etree.SubElement(addr_group, "entry", {"name": token})
                static = etree.SubElement(entry, "static")
                count = 0
                for ip in ips:
                    member = etree.SubElement(static, "member")
                    member.text = f'{ip.parent_token}_{count}'
                    count += 1

        vsys_entry.append(etree.Comment(" Addresses "))
        addr = etree.SubElement(vsys_entry, "address")
        if not no_addr_obj:
            for _, token, ips, _ in self.addressbook.Walk(''):
                count = 0
                for ip in ips:
                    entry = etree.SubElement(addr, "entry", {"name": f'{token}_{count}'})
                    desc = etree.SubElement(entry, "description")
                    desc.text = f'{token}_{count}'
                    elem = etree.SubElement(entry, "ip-netmask")
                    elem.text = str(ip)
                    count += 1

        if add_any_ipv4:
            entry = etree.SubElement(addr, "entry", {"name": "any-ipv4"})
            desc = etree.SubElement(entry, "description")
            desc.text = (
                "Object to match all IPv4 addresses; " "negate to match all IPv6 addresses."
            )
            range = etree.SubElement(entry, "ip-range")
            range.text = ANY_IPV4_RANGE

        vsys_entry.append(tag)

        self.config = config
        document = etree.tostring(config, encoding="UTF-8")
        dom = minidom.parseString(document.decode("UTF-8"))

        return dom.toprettyxml(indent=self.INDENT)



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Common library for network ports and protocol handling."""


class Error(Exception):
    """Base error class."""


class BadPortValue(Error):
    """Invalid port format."""


class BadPortRange(Error):
    """Out of bounds port range."""


class InvalidRange(Error):
    """Range is not valid (eg, single port)."""


class NotSinglePort(Error):
    """Port range defined instead of a single port."""


class PPP:
    """PPP: [P]ort [P]rotocol [P]airs.

    Make port/protocol pairs an object for easy comparisons
    """

    def __init__(self, service) -> None:
        """Init for PPP object.

        Args:
          service: A port/protocol pair as str (eg: '80/tcp', '22-23/tcp') or
                   a nested service name (eg: 'SSH')
        """
        # remove comments (if any)
        self.service = service.split('#')[0].strip()
        if '/' in self.service:
            self.port = self.service.split('/')[0]
            self.protocol = self.service.split('/')[1]
            self.nested = False
        else:
            # for nested services
            self.nested = True
            self.port = None
            self.protocol = None

    @property
    def is_range(self):
        if self.port:
            return '-' in self.port
        else:
            return False

    @property
    def is_single_port(self):
        if self.port:
            return '-' not in self.port
        else:
            return False

    @property
    def start(self):
        # return the first port in the range as int
        if '-' in self.port:
            self._start = int(self.port.split('-')[0])
        else:
            raise InvalidRange(f'{self.port} is not a valid port range')
        return self._start

    @property
    def end(self):
        # return the last port in the range as int
        if '-' in self.port:
            self._end = int(self.port.split('-')[1])
        else:
            raise InvalidRange(f'{self.port} is not a valid port range')
        return self._end

    def __contains__(self, other):
        # determine if a single-port object is within another objects' range
        try:
            return (
                int(self.start) <= int(other.port) <= int(self.end)
            ) and self.protocol == other.protocol
        except:
            raise InvalidRange(f'{self.port} must be a range')

    def __lt__(self, other):
        if self.is_single_port:
            try:
                return int(self.port) < int(other.port)
            except:
                return False
        else:
            raise NotSinglePort('Comparisons cannot be performed on port ranges')

    def __gt__(self, other):
        if self.is_single_port:
            try:
                return int(self.port) > int(other.port)
            except:
                return False
        else:
            raise NotSinglePort('Comparisons cannot be performed on port ranges')

    def __le__(self, other):
        if self.is_single_port:
            try:
                return int(self.port) <= int(other.port)
            except:
                return False
        else:
            raise NotSinglePort('Comparisons cannot be performed on port ranges')

    def __ge__(self, other):
        if self.is_single_port:
            try:
                return int(self.port) >= int(other.port)
            except:
                return False
        else:
            raise NotSinglePort('Comparisons cannot be performed on port ranges')

    def __eq__(self, other):
        if self.is_single_port:
            try:
                return int(self.port) == int(other.port) and self.protocol == other.protocol
            except:
                return False
        else:
            raise NotSinglePort('Comparisons cannot be performed on port ranges')


def Port(port):
    """Sanitize a port value.

    Args:
      port: a port value

    Returns:
      port: a port value

    Raises:
      BadPortValue: port is not valid integer or string
      BadPortRange: port is outside valid range
    """
    pval = -1
    try:
        pval = int(port)
    except ValueError:
        raise BadPortValue(f'port {port} is not valid.')
    if pval < 0 or pval > 65535:
        raise BadPortRange(f'port {port} is out of range 0-65535.')
    return pval
